import gc
import os
import re
import signal
import sys
import warnings
import weakref
from contextlib import ExitStack
from itertools import chain
from pathlib import Path
from shutil import which
from subprocess import CalledProcessError
from subprocess import DEVNULL
from subprocess import PIPE
from subprocess import run

import pytest

from lektor.publisher import _CompatURLStr
from lektor.publisher import _ssh_command
from lektor.publisher import _ssh_key_file
from lektor.publisher import Command
from lektor.publisher import GithubPagesPublisher
from lektor.publisher import GitRepo
from lektor.publisher import publish
from lektor.publisher import Publisher
from lektor.publisher import PublishError
from lektor.utils import locate_executable


def test_ssh_key_file():
    credentials = {"key_file": "/some/id", "key": "rsa:ignored"}
    with _ssh_key_file(credentials) as key_file:
        assert key_file == "/some/id"


def test_ssh_key_file_returns_none():
    with _ssh_key_file({}) as key_file:
        assert key_file is None


@pytest.mark.parametrize("prefix, key_type", [("test:", "TEST"), ("", "RSA")])
def test_ssh_key_file_creates_key(prefix, key_type):
    credentials = {"key": prefix + "".join(["1234567890abcdef"] * 7)}
    with _ssh_key_file(credentials) as key_file:
        assert (
            Path(key_file).read_text()  # pylint: disable=unspecified-encoding
            == (
                f"-----BEGIN {key_type} PRIVATE KEY-----\n"
                "1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef\n"
                "1234567890abcdef1234567890abcdef1234567890abcdef\n"
                f"-----END {key_type} PRIVATE KEY-----\n"
            )
        )
    assert not Path(key_file).exists()


@pytest.mark.parametrize(
    "credentials, port, expected",
    [
        ({}, None, None),
        ({}, 42, "ssh -p 42"),
        ({"key_file": "/keyfile"}, None, 'ssh -i "/keyfile" -o IdentitiesOnly=yes'),
    ],
)
def test_ssh_command(credentials, port, expected):
    with _ssh_command(credentials, port) as key_file:
        assert key_file == expected


def _pyscript(script):
    return (sys.executable, "-c", script)


@pytest.mark.parametrize(
    "cmd, input, expect_rc, expected",
    [
        ("print('xxx')", None, 0, ["xxx"]),
        ("print('yyy', file=sys.stderr)", None, 0, ["yyy"]),
        ("exit(42)", None, 42, []),
        (
            "for _ in sorted(sys.stdin): print(_.strip())",
            "foo\nbar\n",
            0,
            ["bar", "foo"],
        ),
    ],
)
def test_Command_capture(cmd, input, expect_rc, expected):
    command = Command(_pyscript("import sys\n" + cmd), input=input)

    def iter_output():
        proc = yield from command
        assert proc.returncode == expect_rc
        assert proc.stdout is None

    assert list(iter_output()) == expected


def test_Command_env():
    command = Command(
        _pyscript("import os; print(os.environ['TESTVAR'])"),
        env={"TESTVAR": "boo"},
    )
    assert list(command) == ["boo"]


def test_Command_raises_CalledProcessError():
    command = Command(_pyscript("print('x'); exit(42)"), check=True)

    def iter_output():
        with pytest.raises(CalledProcessError) as exc_info:
            yield from command
        assert exc_info.value.returncode == 42

    assert list(iter_output()) == ["x"]


@pytest.mark.parametrize(
    "capture_stdout, silent, stdout, out_err",
    [
        (False, False, None, ("foo\n", "bar\n")),
        (True, False, "foo\n", ("", "bar\n")),
        (False, True, None, ("", "")),
        (True, True, "foo\n", ("", "")),
    ],
)
def test_Command_no_capture(capture_stdout, silent, stdout, out_err, capfd):
    command = Command(
        _pyscript("import sys; print('foo'); print('bar', file=sys.stderr)"),
        capture=False,
        silent=silent,
        capture_stdout=capture_stdout,
    )
    assert command.wait() == 0
    assert command.result().stdout == stdout
    captured = tuple(_.replace(os.linesep, "\n") for _ in capfd.readouterr())
    assert captured == out_err


def test_Command_iter_raises_if_not_capturing():
    command = Command(_pyscript("exit(0)"), capture=False)
    with pytest.raises(RuntimeError) as exc_info:
        next(iter(command))
    assert "Not capturing" in str(exc_info.value)
    assert command.wait() == 0


@pytest.mark.skipif(os.name == "nt", reason="Windows")
# XXX: Can not figure out how to reliably test KeyboardInterrupt handling under Windows.
# Response to signal.SIGINT and/or signal.CTRL_C_EVENT varies depending on what shell
# the test is run under.
def test_Command_handles_keyboard_interrupt():
    pid = os.getpid()
    sig = signal.SIGINT
    with pytest.raises(KeyboardInterrupt):
        command = Command(
            (sys.executable, "-c", f"import os; os.kill({pid}, {sig})"),
            capture_stdout=True,
        )
        list(command)


def test_Command_returncode():
    command = Command((sys.executable, "-c", "exit(42)"))
    assert command.returncode is None
    list(command)
    assert command.returncode == 42


def test_Command_output():
    command = Command(_pyscript("print('foo')"))
    assert list(command.output) == ["foo"]


def test_Command_triggers_no_warnings():
    # This exercises the issue where publishing via rsync resulted
    # in ResourceWarnings about unclosed streams.

    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # This is essentially how RsyncPublisher runs rsync.
        with Command(_pyscript("print()")) as client:
            for _ in client:
                pass

        # The ResourceWarnings regarding unclosed files we are checking for
        # are issued during finalization.  Without this extra effort,
        # finalization wouldn't happen until after the test completes.
        client_is_alive = weakref.ref(client)
        del client
        if client_is_alive():
            gc.collect()

    if client_is_alive():
        warnings.warn(  # noqa: B028
            "Unable to trigger garbage collection of Command instance, "
            "so unable to check for warnings issued during finalization."
        )


@pytest.mark.skipif(
    which("rsync") is None, reason="rsync is not available on this system"
)
@pytest.mark.parametrize("delete", ["yes", "no"])
def test_RsyncPublisher_integration(env, tmp_path, delete):
    # Integration test of local rsync deployment
    # Ensures that RsyncPublisher can successfully invoke rsync
    files = {"file.txt": "content\n"}
    output = tmp_path / "output"
    output.mkdir()
    for path, content in files.items():
        output.joinpath(path).write_text(content)

    target_path = tmp_path / "target"
    target_path.mkdir()
    target = f"rsync://{target_path.resolve()}?delete={delete}"

    event_iter = publish(env, target, output)
    for line in event_iter:
        print(line)

    target_files = {
        os.fspath(_.relative_to(target_path)): _.read_text()
        for _ in target_path.iterdir()
    }
    assert target_files == files


@pytest.fixture
def work_tree(tmp_path):
    work_tree = tmp_path / "work_tree"
    work_tree.mkdir()
    return work_tree


@pytest.fixture
def clean_git_environ(monkeypatch):
    """Delete any GIT_ configuration in the current environment.

    This unsets any existing GIT_SSH_COMMAND, among other things.

    Of note, if one runs tests using git rebase exec, git sets all kinds
    of GIT_* environment variables.
    """
    for key in list(os.environ):
        if key.startswith("GIT_"):
            monkeypatch.delitem(os.environ, key)


@pytest.fixture
def gitrepo(work_tree, clean_git_environ):
    if not locate_executable("git"):
        pytest.skip("no git")
    with GitRepo(work_tree) as repo:
        yield repo


def test_GitRepo_popen(gitrepo):
    assert "No commits yet" in list(gitrepo.popen("status"))


def test_GitRepo_run(gitrepo):
    proc = gitrepo.run("status", capture_stdout=True)
    assert "No commits yet" in proc.stdout


@pytest.mark.usefixtures("clean_git_environ")
def test_GitRepo_set_ssh_credentials(gitrepo):
    gitrepo.set_ssh_credentials({"key_file": "/id_foo"})
    assert gitrepo.environ["GIT_SSH_COMMAND"].startswith('ssh -i "/id_foo"')


@pytest.mark.usefixtures("clean_git_environ")
def test_GitRepo_set_ssh_credentials_no_creds(gitrepo):
    gitrepo.set_ssh_credentials({})
    assert "GIT_SSH_COMMAND" not in gitrepo.environ


def test_GitRepo_set_https_credentials(gitrepo):
    gitrepo.set_https_credentials({"username": "user", "password": "pw"})
    with gitrepo.popen("config", "--local", "--list") as output:
        for line in output:
            m = re.match(r'credential\.helper=store\s+--file\s+"(\S+)"', line)
            if m:
                cred_file = m.group(1)
                break
    # pylint: disable=unspecified-encoding
    assert Path(cred_file).read_text().rstrip() == "https://user:pw@github.com"


def test_GitRepo_set_https_credentials_no_cred(gitrepo):
    gitrepo.set_https_credentials({})
    assert not any(
        line.startswith("credential.helper")
        for line in gitrepo.popen("config", "--local", "--list")
    )


def test_GitRepo_add_to_index(gitrepo):
    gitrepo.add_to_index("testfile", "testcontent\n")
    proc = gitrepo.run("diff", "--cached", capture_stdout=True)
    assert "+testcontent" in proc.stdout


class DummyUpstreamRepo:
    def __init__(self, git_dir):
        run(("git", "init", "--bare", git_dir), check=True, stdout=DEVNULL)
        self.git_dir = git_dir

    @property
    def url(self):
        return self.git_dir

    def run(self, args, **kwargs):
        cmd = ("git", "--bare", "--git-dir", self.git_dir) + args
        # XXX: Use text=True instead of encoding for py>3.6
        return run(cmd, stdout=PIPE, encoding="utf-8", check=False)

    def count_commits(self, branch):
        proc = self.run(("log", "--pretty=oneline", branch))
        return len(proc.stdout.splitlines())

    def ls_files(self, treeish):
        return set(self.iter_files(treeish))

    def iter_files(self, treeish, prefix=""):
        # Currently only lists top-level files
        for line in self.run(("ls-tree", treeish)).stdout.splitlines():
            _, typ, oid, name = line.split(maxsplit=3)
            if typ == "blob":
                yield prefix + name
            else:
                assert typ == "tree"
                yield from self.iter_files(oid, prefix=f"{prefix}{name}/")


@pytest.fixture
def upstream_repo(tmp_path, clean_git_environ):
    git_dir = tmp_path / "upstream.git"
    return DummyUpstreamRepo(git_dir.__fspath__())


@pytest.fixture
def no_scary_output(capsys):
    # Check that no scary messages are output to stdout
    def is_scary(line):
        return line.startswith("fatal")

    yield
    captured = capsys.readouterr()
    scary_output = list(
        filter(is_scary, chain(captured.out.splitlines(), captured.err.splitlines()))
    )
    assert len(scary_output) == 0


@pytest.fixture
def publish_ghpages(work_tree, clean_git_environ, no_scary_output):
    # Run publish_ghpages on a fresh GitRepo instance
    if not locate_executable("git"):
        pytest.skip("no git")

    def GitRepo_publish_ghpages(*args, **kwargs):
        with GitRepo(work_tree) as repo:
            rv = yield from repo.publish_ghpages(*args, **kwargs)
            return rv

    return GitRepo_publish_ghpages


def test_GitRepo_publish_ghpages(publish_ghpages, upstream_repo, work_tree):
    test_file = work_tree / "new-file"
    test_file.write_text("test\n")
    for line in publish_ghpages(upstream_repo.url, "gh-pages"):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 1
    assert upstream_repo.ls_files("gh-pages") == {"new-file"}

    test_file.rename(work_tree / "renamed-file")
    for line in publish_ghpages(upstream_repo.url, "gh-pages"):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 2
    assert upstream_repo.ls_files("gh-pages") == {"renamed-file"}


def test_GitRepo_publish_ghpages_cname(publish_ghpages, upstream_repo, work_tree):
    for line in publish_ghpages(upstream_repo.url, "gh-pages", "example.org"):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 1
    assert upstream_repo.ls_files("gh-pages") == {"CNAME"}

    for line in publish_ghpages(upstream_repo.url, "gh-pages"):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 2
    assert upstream_repo.ls_files("gh-pages") == set()


def test_GitRepo_publish_ghpages_ignores_lektor_dir(
    publish_ghpages, upstream_repo, work_tree
):
    lektor_dir = work_tree / ".lektor"
    lektor_dir.mkdir()
    lektor_dir.joinpath("ignored").write_text("should not be published")
    work_tree.joinpath("included").write_text("should be published")

    for line in publish_ghpages(upstream_repo.url, "gh-pages"):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 1
    assert upstream_repo.ls_files("gh-pages") == {"included"}


def test_GitRepo_publish_ghpages_no_changes(publish_ghpages, upstream_repo):
    for line in publish_ghpages(upstream_repo.url, "gh-pages", "example.com"):
        print(line)
    assert upstream_repo.ls_files("gh-pages") == {"CNAME"}
    assert any(
        "No changes" in line
        for line in publish_ghpages(upstream_repo.url, "gh-pages", "example.com")
    )


def test_GitRepo_publish_ghpages_discard_history(
    publish_ghpages, upstream_repo, work_tree
):
    for line in publish_ghpages(upstream_repo.url, "gh-pages", "example.com"):
        print(line)
    assert upstream_repo.ls_files("gh-pages") == {"CNAME"}

    work_tree.joinpath("myfile").write_text("contents")
    for line in publish_ghpages(upstream_repo.url, "gh-pages", None, False):
        print(line)
    assert upstream_repo.count_commits("gh-pages") == 1
    assert upstream_repo.ls_files("gh-pages") == {"myfile"}


@pytest.fixture
def output_path(tmp_path):
    return str(tmp_path / "output_path")


@pytest.fixture
def ghp_publisher(env, output_path):
    return GithubPagesPublisher(env, output_path)


@pytest.mark.parametrize(
    "target_url, publish_args, warns",
    [
        (
            "ghpages://owner/project?cname=example.com&preserve_history=no",
            (
                "ssh://git@github.com/owner/project.git",
                "gh-pages",
                "example.com",
                False,
            ),
            False,
        ),
        (
            "ghpages+https://owner/owner.github.io",
            ("https://github.com/owner/owner.github.io.git", "master", None, True),
            True,
        ),
    ],
)
@pytest.mark.skipif(not locate_executable("git"), reason="no git")
def test_GithubPagesPublisher_publish(
    ghp_publisher, output_path, mocker, target_url, publish_args, warns
):
    GitRepo = mocker.patch("lektor.publisher.GitRepo", spec_set=True)
    repo = mocker.Mock(spec_set=dir(GitRepo))
    repo.publish_ghpages.return_value = iter(["Published!"])
    GitRepo.return_value.__enter__.return_value = repo

    with ExitStack() as stack:
        if warns:
            stack.enter_context(pytest.deprecated_call())
        output = list(ghp_publisher.publish(target_url))

    GitRepo.assert_called_once_with(output_path)
    if "+https:" in target_url:
        repo.set_https_credentials.assert_called_once_with({})
    else:
        repo.set_ssh_credentials.assert_called_once_with({})
    repo.publish_ghpages.assert_called_once_with(*publish_args)
    if warns:
        assert "WARNING" in " ".join(output)
        output = output[-1:]
    assert output == ["Published!"]


@pytest.mark.usefixtures("no_utils")
def test_GithubPagesPublisher_publish_fails_if_no_git(ghp_publisher):
    target_url = "ghpages://owner/project"
    with pytest.raises(PublishError) as exc_info:
        list(ghp_publisher.publish(target_url))
    assert re.search(r"git.*not found", str(exc_info.value))


@pytest.mark.parametrize(
    "target_url, expected",
    [
        (
            "ghpages://owner/project",
            ("ssh://git@github.com/owner/project.git", "gh-pages", None, True, []),
        ),
        (
            "ghpages://owner/owner.github.io?branch=main",
            ("ssh://git@github.com/owner/owner.github.io.git", "main", None, True, []),
        ),
        (
            "ghpages://owner/project?branch=brnch&cname=example.com",
            (
                "ssh://git@github.com/owner/project.git",
                "brnch",
                "example.com",
                True,
                [],
            ),
        ),
        (
            "ghpages://owner/project?preserve_history=no",
            ("ssh://git@github.com/owner/project.git", "gh-pages", None, False, []),
        ),
        (
            "ghpages://owner/project?preserve_history=yes",
            ("ssh://git@github.com/owner/project.git", "gh-pages", None, True, []),
        ),
        (
            "ghpages+ssh://own/proj",
            ("ssh://git@github.com/own/proj.git", "gh-pages", None, True, []),
        ),
        (
            "ghpages+https://own/proj",
            ("https://github.com/own/proj.git", "gh-pages", None, True, []),
        ),
        (
            "ghpages+ssh://own:22/proj",
            ("ssh://git@github.com/own/proj.git", "gh-pages", None, True, []),
        ),
        (
            "ghpages+https://own:443/proj",
            ("https://github.com/own/proj.git", "gh-pages", None, True, []),
        ),
    ],
)
def test_GithubPagesPublisher_parse_url(ghp_publisher, target_url, expected):
    assert ghp_publisher._parse_url(target_url) == expected


def test_GithubPagesPublisher_parse_url_warns_on_default_master_branch(ghp_publisher):
    target_url = "ghpages://owner/owner.github.io"
    with pytest.deprecated_call():
        push_url, branch, cname, preserve_history, warnings = ghp_publisher._parse_url(
            target_url
        )
    assert (push_url, branch, cname, preserve_history) == (
        "ssh://git@github.com/owner/owner.github.io.git",
        "master",
        None,
        True,
    )
    assert len(warnings) > 0
    assert re.search(r"WARNING.*set.*branch", " ".join(warnings))


@pytest.mark.parametrize(
    "target_url, expected",
    [
        ("ghpages:///project", "owner missing"),
        ("ghpages://owner", "project missing"),
        ("ghpages://owner:2222/project", "non-standard port"),
        ("ghpages+https://owner:4430/project", "non-standard port"),
    ],
)
def test_GithubPagesPublisher_parse_url_failures(ghp_publisher, target_url, expected):
    with pytest.raises(PublishError) as exc_info:
        ghp_publisher._parse_url(target_url)
    assert expected in str(exc_info.value)


@pytest.mark.parametrize(
    "target_url, credentials, expected",
    [
        ("ghpages://owner/proj", None, {}),
        (
            "ghpages://owner/proj",
            {"username": "user", "password": "pw"},
            {"username": "user", "password": "pw"},
        ),
        ("ghpages://user:pw@owner/proj", None, {"username": "user", "password": "pw"}),
        (
            "ghpages://user:pw@owner/proj",
            {"arbi": "trary"},
            {"username": "user", "password": "pw", "arbi": "trary"},
        ),
        ("ghpages://user:pw@owner/proj", None, {"username": "user", "password": "pw"}),
        (
            "ghpages://user:pw@owner/proj",
            {"password": "secret"},
            {"username": "user", "password": "secret"},
        ),
    ],
)
def test_GithubPagesPublisher_parse_credentials(
    ghp_publisher, credentials, target_url, expected
):
    assert ghp_publisher._parse_credentials(credentials, target_url) == expected


def test_publish(env, output_path, mocker):
    Publisher = mocker.Mock()
    env.add_publisher("publishtest", Publisher)
    credentials = {"foo": "bar"}
    rv = publish(env, "publishtest://host/path", output_path, credentials)
    target_url = "publishtest://host/path"
    assert Publisher.mock_calls == [
        mocker.call(env, output_path),
        mocker.call().publish(target_url, credentials),
    ]
    assert rv is Publisher.return_value.publish.return_value


def test_publish_unknown_scheme(env, output_path):
    with pytest.raises(PublishError) as exc_info:
        publish(env, "unknownscheme://host/path", output_path)
    assert "unknown scheme" in str(exc_info.value)


def test_publish_old_plugin(env, output_path, mocker):
    target_url = "publishtest://Bücher.example/?foo=bar"

    class OldPublisher(Publisher):
        # pylint: disable=signature-differs
        def publish(self, target_url, credentials, **extra):
            return target_url

    env.add_publisher("publishtest", OldPublisher)
    compat_url = publish(env, target_url, output_path, credentials={})

    assert isinstance(compat_url, _CompatURLStr)


def test_CompatURLStr_is_str():
    url = _CompatURLStr(" scheme://example.org/ ")
    assert isinstance(url, str)
    assert str(url) == " scheme://example.org/ "


def test_CompatURLStr_raises_attribute_error():
    url = _CompatURLStr("scheme://example.org/")
    with pytest.raises(AttributeError):
        url._replace()


def test_CompatURLStr_works_as_URL():
    url = _CompatURLStr("scheme:///path?foo=bar")
    with pytest.deprecated_call() as w:
        warnings.filterwarnings("ignore", "'werkzeug", DeprecationWarning)
        assert dict(url.decode_query()) == {"foo": "bar"}
    assert w[0].filename == __file__



from warnings import warn

from lektor.markdown import Markdown
from lektor.types.base import Type


class MarkdownDescriptor:
    def __init__(self, source, options):
        self.source = source
        self.options = options

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return Markdown(self.source, record=obj, field_options=self.options)


class MarkdownType(Type):
    widget = "multiline-text"

    def __init__(self, env, options):
        super().__init__(env, options)
        _check_option(options, "resolve_links", ("always", "never", "when-possible"))

    def value_from_raw(self, raw):
        return MarkdownDescriptor(raw.value or "", self.options)


def _check_option(options, name, choices):
    value = options.get(name)
    if value is not None and value not in choices:
        message = (
            f"Unrecognized value {value!r} for the {name!r} markdown field option. "
            f"Valid values are: {', '.join(repr(_) for _ in choices)}."
        )
        warn(message, stacklevel=3)



"""Test for issue `#410`_.

This tests for at least one of the bugs reported in `#410`_, namely
 that paginated pages were not being built if they had no children.

.. _#410: https://github.com/lektor/lektor/issues/410

"""

import os

import pytest


@pytest.fixture(
    params=[
        pytest.param(True, id="paginated"),
        pytest.param(False, id="non-paginated"),
    ]
)
def scratch_project_data(scratch_project_data, request):
    # Specialize the inherited scratch project (from conftest.py)
    # by (possibly) enabling pagination for models/page.ini.
    is_paginated = request.param
    page_ini = scratch_project_data / "models/page.ini"
    if is_paginated:
        page_ini.write_text(
            "\n".join(
                [
                    page_ini.read_text("utf-8"),
                    "[pagination]",
                    "enabled = yes",
                    "",
                ]
            ),
            "utf-8",
        )
    return scratch_project_data


def test_build_childless_page(scratch_builder):
    # Test that a basic childless page gets built (whether it is
    # paginated or not)
    scratch_builder.build_all()
    index_html = os.path.join(scratch_builder.destination_path, "index.html")
    assert os.path.exists(index_html)



import pytest

from lektor.build_programs import BuildError
from lektor.builder import Builder
from lektor.project import Project
from lektor.reporter import BufferReporter


@pytest.fixture
def pad(data_path, save_sys_path):
    proj = Project.from_path(data_path / "ünicöde-project")
    return proj.make_env().new_pad()


@pytest.fixture
def builder(pad, tmp_path):
    output_path = tmp_path / "output"
    output_path.mkdir()
    return Builder(pad, str(output_path))


def test_unicode_project_folder(pad, builder):
    prog, _ = builder.build(pad.root)
    with prog.artifacts[0].open("rb") as f:
        assert f.read() == b"<h1>Hello</h1>\n<p>W\xc3\xb6rld</p>\n\n"


def test_unicode_attachment_filename(pad, builder):
    with BufferReporter(builder.env) as reporter:
        prog, _ = builder.build(pad.root.attachments.first())

        failures = reporter.get_failures()
        assert len(failures) == 0

        with prog.artifacts[0].open("rb") as f:
            assert f.read().rstrip() == b"attachment"


def test_bad_file_ignored(pad, builder):
    record = pad.root.children.first()
    with BufferReporter(builder.env) as reporter:
        builder.build(record)
        failures = reporter.get_failures()
        assert len(failures) == 1
        exc_info = failures[0]["exc_info"]
        assert exc_info[0] is BuildError
        assert (
            "The URL for this record contains non "
            "ASCII characters" in exc_info[1].message
        )



from __future__ import annotations

import fnmatch
import os
import uuid
from functools import update_wrapper
from typing import TYPE_CHECKING

import babel.dates
import jinja2
from jinja2.loaders import split_template_path

from lektor.constants import PRIMARY_ALT
from lektor.context import config_proxy
from lektor.context import get_asset_url
from lektor.context import get_ctx
from lektor.context import get_locale
from lektor.context import site_proxy
from lektor.context import url_to
from lektor.environment.config import Config
from lektor.environment.config import DEFAULT_CONFIG  # noqa - reexport
from lektor.environment.config import ServerInfo  # noqa - reexport
from lektor.environment.config import update_config_from_ini  # noqa - reexport
from lektor.environment.expressions import Expression  # noqa - reexport
from lektor.environment.expressions import FormatExpression  # noqa - reexport
from lektor.markdown import Markdown
from lektor.packages import load_packages
from lektor.pluginsystem import initialize_plugins
from lektor.pluginsystem import PluginController
from lektor.publisher import builtin_publishers
from lektor.utils import format_lat_long
from lektor.utils import tojson_filter


if TYPE_CHECKING:
    from typing import Literal

    from lektor.assets import Asset
    from lektor.build_programs import BuildProgram
    from lektor.sourceobj import SourceObject


def _prevent_inlining(wrapped):
    """Ensure wrapped jinja filter does not get inlined by the template compiler.

    The jinja compiler normally assumes that filters are pure functions (whose
    result depends only on their parameters) and will inline filter calls that
    are applied to compile-time constants.

    E.g.

        'say {{ "foo" | upper }}'

    will be compiled to

        "say Foo"

    Many of our filters depend on global state (e..g the Lektor build context).

    Applying this decorator to them will ensure they are not inlined.
    """

    # the use of @pass_context will prevent inlining
    @jinja2.pass_context
    def wrapper(_jinja_ctx, *args, **kwargs):
        return wrapped(*args, **kwargs)

    return update_wrapper(wrapper, wrapped)


def _dates_filter(name, wrapped):
    """Wrap one of the babel.dates.format_* functions for use as a jinja filter.

    This will create a jinja filter that will:

    - Check for *undefined* date/time input (and, in that case, return an empty string).

    - Check that the ``format`` and ``locale`` parameters, if provided, have the correct
      types, otherwise raising ``TypeError``.

    - Raise ``TypeError`` with a somewhat informative message if the wrapped formatting
      function raises an unexpected exception.  Such an exception is most likely due to
      being passed an unsupported date/time time.  (The Babel formatting functions
      accept a fairly wide range of input types — and that range might potentially vary
      between releases — so we do not explicitly check the input type before passing it
      on to Babel.)

    If `locale` is not specified, we fill it in based on the current *alt*.

    """

    @_prevent_inlining
    def wrapper(arg, format="medium", **kwargs):
        if isinstance(arg, jinja2.Undefined):
            # This will typically return an empty string, though it depends on the
            # specific type of undefined instance.  E.g. if arg is a DebugUndefined, it
            # will return a more descriptive message, and if arg is a StrictUndefined,
            # an UndefinedError will be raised.
            return str(arg)

        if not isinstance(format, str):
            raise TypeError(
                f"The 'format' parameter to '{name}' should be a str, not {format!r}"
            )

        locale = kwargs.get("locale")
        if locale is None:
            kwargs["locale"] = get_locale("en_US")

        try:
            return wrapped(arg, format, **kwargs)
        except (TypeError, ValueError):
            raise
        except Exception as exc:
            raise TypeError(
                f"While evaluating filter '{name}', an unexpected exception was "
                "raised. This is likely caused by an input or parameter of an "
                "unsupported type."
            ) from exc

    return update_wrapper(wrapper, wrapped)


@_prevent_inlining
def _markdown_filter(
    source: str,
    *,
    resolve_links: Literal["always", "never", "when-possible", None] = None,
    **kw: str,
) -> Markdown:
    """A jinja filter that converts markdown text to HTML."""
    ctx = get_ctx()
    source_obj = ctx.source if ctx is not None else None
    return Markdown(
        source, source_obj, field_options={**kw, "resolve_links": resolve_links}
    )


# Special files that should always be ignored.
IGNORED_FILES = ["thumbs.db", "desktop.ini", "Icon\r"]

# These files are important for artifacts and must not be ignored when
# they are built even though they start with dots.
SPECIAL_ARTIFACTS = [".htaccess", ".htpasswd"]

# Default glob pattern of ignored files.
EXCLUDED_ASSETS = ["_*", ".*"]

# Default glob pattern of included files (higher-priority than EXCLUDED_ASSETS).
INCLUDED_ASSETS = []


def any_fnmatch(filename, patterns):
    for pat in patterns:
        if fnmatch.fnmatch(filename, pat):
            return True

    return False


class CustomJinjaEnvironment(jinja2.Environment):
    def _load_template(self, name, globals):
        ctx = get_ctx()

        try:
            rv = jinja2.Environment._load_template(self, name, globals)
            if ctx is not None:
                filename = rv.filename
                ctx.record_dependency(filename)
            return rv
        except jinja2.TemplateSyntaxError as e:
            if ctx is not None:
                ctx.record_dependency(e.filename)
            raise
        except jinja2.TemplateNotFound as e:
            if ctx is not None:
                # If we can't find the template we want to record at what
                # possible locations the template could exist.  This will help
                # out watcher to pick up templates that will appear in the
                # future.  This assumes the loader is a file system loader.
                for template_name in e.templates:
                    pieces = split_template_path(template_name)
                    for base in self.loader.searchpath:
                        ctx.record_dependency(os.path.join(base, *pieces))
            raise


@jinja2.pass_context
def lookup_from_bag(jinja_ctx, *args):
    pieces = ".".join(x for x in args if x)
    site = jinja_ctx.get("site", default=site_proxy)
    return site.databags.lookup(pieces)


class Environment:
    def __init__(self, project, load_plugins=True, extra_flags=None):
        self.project = project
        self.root_path = os.path.abspath(project.tree)

        self.theme_paths = [
            os.path.join(self.root_path, "themes", theme)
            for theme in self.project.themes
        ]

        if not self.theme_paths:
            # load the directories in the themes directory as the themes
            try:
                for fname in os.listdir(os.path.join(self.root_path, "themes")):
                    f = os.path.join(self.root_path, "themes", fname)
                    if os.path.isdir(f):
                        self.theme_paths.append(f)
            except OSError:
                pass

        template_paths = [
            os.path.join(path, "templates")
            for path in [self.root_path] + self.theme_paths
        ]

        self.jinja_env = CustomJinjaEnvironment(
            autoescape=self.select_jinja_autoescape,
            extensions=["jinja2.ext.do"],
            loader=jinja2.FileSystemLoader(template_paths),
        )

        from lektor.db import F  # pylint: disable=import-outside-toplevel
        from lektor.db import get_alts  # pylint: disable=import-outside-toplevel

        def latlongformat(latlong, secs=True):
            lat, lon = latlong
            return format_lat_long(lat=lat, long=lon, secs=secs)

        self.jinja_env.filters.update(
            tojson=tojson_filter,
            latformat=lambda x, secs=True: format_lat_long(lat=x, secs=secs),
            longformat=lambda x, secs=True: format_lat_long(long=x, secs=secs),
            latlongformat=latlongformat,
            url=_prevent_inlining(url_to),
            asseturl=_prevent_inlining(get_asset_url),
            markdown=_markdown_filter,
        )
        self.jinja_env.globals.update(
            F=F,
            url_to=url_to,
            site=site_proxy,
            config=config_proxy,
            bag=lookup_from_bag,
            get_alts=get_alts,
            get_random_id=lambda: uuid.uuid4().hex,
        )
        self.jinja_env.filters.update(
            dateformat=_dates_filter("dateformat", babel.dates.format_date),
            datetimeformat=_dates_filter("datetimeformat", babel.dates.format_datetime),
            timeformat=_dates_filter("timeformat", babel.dates.format_time),
        )

        # pylint: disable=import-outside-toplevel
        from lektor.types import builtin_types

        self.types = builtin_types.copy()

        self.publishers = builtin_publishers.copy()

        # The plugins that are loaded for this environment.  This is
        # modified by the plugin controller and registry methods on the
        # environment.
        self.plugin_controller = PluginController(self, extra_flags)
        self.plugins = {}
        self.plugin_ids_by_class = {}
        self.build_programs = []
        self.special_file_assets = {}
        self.special_file_suffixes = {}
        self.custom_url_resolvers = []
        self.custom_generators = []
        self.virtual_sources = {}

        if load_plugins:
            self.load_plugins()
        # pylint: disable=import-outside-toplevel
        from lektor.db import siblings_resolver

        self.virtualpathresolver("siblings")(siblings_resolver)

    root_path: str
    build_programs: list[tuple[type[SourceObject], type[BuildProgram]]]
    special_file_assets: dict[str, type[Asset]]
    special_file_suffixes: dict[str, str]

    @property
    def asset_path(self):
        return os.path.join(self.root_path, "assets")

    @property
    def temp_path(self):
        return os.path.join(self.root_path, "temp")

    def load_plugins(self):
        """Loads the plugins."""
        load_packages(self)
        initialize_plugins(self)

    def load_config(self):
        """Loads the current config."""
        return Config(self.project.project_file)

    def new_pad(self):
        """Convenience function to create a database and pad."""
        from lektor.db import Database  # pylint: disable=import-outside-toplevel

        return Database(self).new_pad()

    def is_uninteresting_source_name(self, filename: str) -> bool:
        """These files are ignored when sources are built into artifacts."""
        fn = filename.lower()
        if fn in SPECIAL_ARTIFACTS:
            return False

        proj = self.project
        if any_fnmatch(filename, INCLUDED_ASSETS + proj.included_assets):
            # Included by the user's project config, thus not uninteresting.
            return False
        return any_fnmatch(filename, EXCLUDED_ASSETS + proj.excluded_assets)

    @staticmethod
    def is_ignored_artifact(asset_name):
        """This is used by the prune tool to figure out which files in the
        artifact folder should be ignored.
        """
        fn = asset_name.lower()
        if fn in SPECIAL_ARTIFACTS:
            return False
        return fn[:1] == "." or fn in IGNORED_FILES

    def render_template(self, name, pad=None, this=None, values=None, alt=None):
        ctx = self.make_default_tmpl_values(pad, this, values, alt, template=name)
        return self.jinja_env.get_or_select_template(name).render(ctx)

    def make_default_tmpl_values(
        self, pad=None, this=None, values=None, alt=None, template=None
    ):
        values = dict(values or ())

        # If not provided, pick the alt from the provided "this" object.
        # As there is no mandatory format for it, we make sure that we can
        # deal with a bad attribute there.
        if alt is None:
            if this is not None:
                alt = getattr(this, "alt", None)
                if not isinstance(alt, str):
                    alt = None
            if alt is None:
                alt = PRIMARY_ALT

        # This is already a global variable but we can inject it as a
        # local override if available.
        if pad is None:
            ctx = get_ctx()
            if ctx is not None:
                pad = ctx.pad
        if pad is not None:
            values["site"] = pad
        if this is not None:
            values["this"] = this
        if alt is not None:
            values["alt"] = alt
        self.plugin_controller.emit(
            "process-template-context", context=values, template=template
        )
        return values

    @staticmethod
    def select_jinja_autoescape(filename):
        if filename is None:
            return False
        return filename.endswith((".html", ".htm", ".xml", ".xhtml"))

    def resolve_custom_url_path(self, obj, url_path):
        for resolver in self.custom_url_resolvers:
            rv = resolver(obj, url_path)
            if rv is not None:
                return rv
        return None

    # -- methods for the plugin system

    def add_build_program(
        self, cls: type[SourceObject], program: type[BuildProgram]
    ) -> None:
        self.build_programs.append((cls, program))

    def add_asset_type(
        self, asset_cls: type[Asset], build_program: type[BuildProgram]
    ) -> None:
        self.build_programs.append((asset_cls, build_program))
        self.special_file_assets[asset_cls.source_extension] = asset_cls
        if asset_cls.artifact_extension:
            cext = asset_cls.source_extension + asset_cls.artifact_extension
            self.special_file_suffixes[cext] = asset_cls.source_extension

    def add_publisher(self, scheme, publisher):
        if scheme in self.publishers:
            raise RuntimeError(f"Scheme {scheme!r} is already registered.")
        self.publishers[scheme] = publisher

    def add_type(self, type):
        name = type.name
        if name in self.types:
            raise RuntimeError(f"Type {name!r} is already registered.")
        self.types[name] = type

    def virtualpathresolver(self, prefix):
        def decorator(func):
            if prefix in self.virtual_sources:
                raise RuntimeError(f"Prefix {prefix!r} is already registered.")
            self.virtual_sources[prefix] = func
            return func

        return decorator

    def urlresolver(self, func):
        self.custom_url_resolvers.append(func)
        return func

    def generator(self, func):
        self.custom_generators.append(func)
        return func



import json
import os
from importlib import resources
from pathlib import Path


translations = {}
for file in resources.files("lektor").joinpath("translations").iterdir():
    filename = Path(file.name)
    if filename.suffix == ".json" and file.is_file():
        lang = filename.stem
        with file.open("rb") as f:
            translations[lang] = json.load(f)

KNOWN_LANGUAGES = set(translations.keys())


def get_translations(language):
    """Looks up the translations for a given language."""
    return translations.get(language)


def is_valid_language(lang):
    """Verifies a language is known and valid."""
    return lang in KNOWN_LANGUAGES


def get_default_lang():
    """Returns the default language the system should use."""
    for key in "LANGUAGE", "LC_ALL", "LC_CTYPE", "LANG":
        value = os.environ.get(key)
        if not value:
            continue
        lang = value.split("_")[0].lower()
        if is_valid_language(lang):
            return lang
    return "en"


def load_i18n_block(key):
    """Looks up an entire i18n block from a known translation."""
    rv = {}
    for lang in KNOWN_LANGUAGES:
        val = translations.get(lang, {}).get(key)
        if val is not None:
            rv[lang] = val
    return rv


def get_i18n_block(inifile_or_dict, key, pop=False):
    """Extracts an i18n block from an ini file or dictionary for a given
    key. If "pop", delete keys from "inifile_or_dict".
    """
    rv = {}
    for k in list(inifile_or_dict):
        if k == key:
            # English is the internal default language with preferred
            # treatment.
            rv["en"] = inifile_or_dict.pop(k) if pop else inifile_or_dict[k]
        elif k.startswith(key + "["):
            rv[k[len(key) + 1 : -1]] = (
                inifile_or_dict.pop(k) if pop else inifile_or_dict[k]
            )
    return rv


def generate_i18n_kvs(**opts):
    """Generates key-value pairs based on the kwargs passed into this function.
    For every key ending in "_i18n", its corresponding value will be translated
    and returned once for every language that has a known translation.
    """
    for key, value in opts.items():
        if key.endswith("_i18n"):
            base_key = key[:-5]
            for lang, trans in load_i18n_block(value).items():
                lang_key = f"{base_key}[{lang}]"
                yield lang_key, trans
        else:
            yield key, value



import re

import pytest

from lektor.context import Context
from lektor.environment import Environment
from lektor.reporter import BufferReporter
from lektor.sourceobj import VirtualSourceObject
from lektor.utils import cleanup_path
from lektor.utils import cleanup_url_path


# pylint: disable-next=wrong-import-order
from conftest import restore_import_state  # isort: skip


@pytest.fixture(scope="module")
def pad(project):
    # Use a module-scoped pad to speed tests
    with restore_import_state():
        return Environment(project).new_pad()


@pytest.fixture
def config(pad, monkeypatch):
    """Patch a function-scoped config into our session-scoped pad.

    (For those tests/fixtures that want to mangle the config.)
    """
    config = pad.db.env.load_config()
    monkeypatch.setattr(pad.db, "config", config)
    return config


@pytest.mark.parametrize(
    "path, expected",
    [
        ("", "/"),
        ("/", "/"),
        ("/foo", "/foo"),
        ("//foo", "/foo"),
        ("///foo", "/foo"),
        ("/foo/", "/foo"),
        ("/////foo/", "/foo"),
        ("/////foo////", "/foo"),
        ("/////foo/.///", "/foo"),
        ("/////foo/..///", "/"),
        (".", "/"),
        ("/foo/./bar/", "/foo/bar"),
        ("/foo/../bar/", "/bar"),
    ],
)
def test_cleanup_path(path, expected):
    assert cleanup_path(path) == expected


@pytest.mark.parametrize(
    "url_path, expected",
    [
        ("", "/"),
        ("foo", "/foo"),
        ("foo/", "/foo"),
        ("/", "/"),
        ("//", "/"),
        ("/foo", "/foo"),
        ("/foo/", "/foo"),
        ("///foo", "/foo"),
        ("////foo", "/foo"),
        ("/////foo", "/foo"),
        ("foo///bar", "/foo/bar"),
        ("./foo/.", "/foo"),
        ("/foo/./bar", "/foo/bar"),
        ("../foo", "/foo"),
        ("/../foo/", "/foo"),
        ("foo/../bar", "/bar"),
        ("foo/../../bar/", "/bar"),
        ("foo#frag", "/foo"),  # fragment gets stripped
        ("/foo?query", "/foo"),  # query get stripped
        ("///foo", "/foo"),  # empty netloc
        ("http:foo", "/foo"),  # explicit scheme
        ("HTTP:FOO", "/FOO"),  # explicit scheme get case normalized
        ("https:///foo", "/foo"),  # explicit scheme
    ],
)
def test_cleanup_url_path(url_path, expected):
    assert cleanup_url_path(url_path) == expected


@pytest.mark.parametrize(
    "url_path",
    [
        "file:///foo",  # bad scheme
        "//example.net/bar",  # netloc not allowed
    ],
)
def test_cleanup_url_path_raises_value_error(url_path):
    with pytest.raises(ValueError):
        cleanup_url_path(url_path)


@pytest.mark.parametrize(
    "from_path, from_alt, to_path, to_alt, expected",
    [
        ("/projects/wolf", "en", "/projects/slave", "en", "../slave/"),
        ("/projects/wolf", "de", "/projects/slave", "de", "../sklave/"),
        ("/projects/slave", "en", "/projects/slave", "de", "../../de/projects/sklave/"),
        ("/projects/slave", "de", "/projects/slave", "en", "../../../projects/slave/"),
        ("/projects/wolf", "en", "/projects/wolf", "en", "./"),
    ],
)
def test_url_to_page(pad, from_path, from_alt, to_path, to_alt, expected):
    from_ = pad.get(from_path, alt=from_alt)
    to = pad.get(to_path, alt=to_alt)
    assert from_.url_to(to) == expected


@pytest.mark.parametrize(
    "alt, expected",
    [
        ("en", "page/2/"),
        ("de", "../de/blog/page/2/"),
    ],
)
def test_url_to_page_with_explicit_alt(pad, alt, expected):
    page1 = pad.get("/blog", alt="en", page_num=1)
    page2 = pad.get("/blog", alt="en", page_num=2)
    assert page1.url_to(page2, alt=alt) == expected


@pytest.fixture
def build_context(builder):
    build_state = builder.new_build_state()
    with Context(build_state.new_artifact("dummy-artifact")) as ctx:
        yield ctx


@pytest.mark.usefixtures("build_context")
def test_url_to_thumbnail(pad):
    extra_de = pad.get("/extra", alt="de")
    thumbnail = pad.get("/test.jpg").thumbnail(42)
    assert extra_de.url_to(thumbnail) == "../../test@42x56.jpg"


@pytest.fixture
def external_url(config):
    config.values["PROJECT"]["url"] = "http://example.org/site/"


@pytest.mark.parametrize(
    "path, kwargs, expected",
    [
        ("/projects/slave", {}, "../slave/"),
        ("/projects/slave", {"absolute": True}, "/site/projects/slave/"),
        (
            "/projects/slave",
            {"external": True},
            "http://example.org/site/projects/slave/",
        ),
        (
            "/projects/slave",
            {"absolute": True, "external": True},
            "/site/projects/slave/",
        ),
    ],
)
@pytest.mark.usefixtures("external_url")
def test_url_to_external(pad, path, kwargs, expected):
    wolf_en = pad.get("/projects/wolf")
    assert wolf_en.url_to(path, **kwargs) == expected


NO_PROJECT_URL = RuntimeError(r"To use absolute.*configure.*URL")
CAN_NOT_RESOLVE = RuntimeError(r"Can not resolve")
NETLOC_NOT_ALLOWED = RuntimeError(r"Netloc not allowed")
CONFLICTING_ALT = RuntimeError(r"Conflicting.*alt")
RESOLVE_INCOMPATIBLE_WITH_BANG = RuntimeError(r"Resolve=True.*incompatible.*!")


def _id_for_dict(value):
    if isinstance(value, dict):
        return ",".join(f"{k}={v!r}" for k, v in value.items())
    return None


@pytest.mark.parametrize(
    "path, kwargs, expected",
    [
        ("", {}, "./"),
        (".", {}, "./"),
        ("./", {}, "./"),
        ("../", {}, "../"),
        ("../wolf", {}, "./"),
        ("../wolf/", {}, "./"),
        ("../wolf/.", {}, "./"),
        ("../slave", {}, "../slave/"),
        ("../wolf/../slave/", {}, "../slave/"),
        ("/", {}, "../../"),
        ("/extra", {}, "../../extra/"),
        ("/projects/wolf", {}, "./"),
        ("/projects/slave", {}, "../slave/"),
        # Test alt
        (".", {"alt": "de"}, "../../de/projects/wolf/"),
        ("../slave", {"alt": "de"}, "../../de/projects/sklave/"),
        ("/projects/wolf", {"alt": "de"}, "../../de/projects/wolf/"),
        ("/projects/slave", {"alt": "de"}, "../../de/projects/sklave/"),
        (".", {"alt": ""}, "./"),
        # Test absolute
        (".", {"absolute": True}, "/projects/wolf/"),
        ("/projects/wolf", {"absolute": True}, "/projects/wolf/"),
        ("../slave", {"absolute": True}, "/projects/slave/"),
        ("/projects/slave", {"absolute": True}, "/projects/slave/"),
        ("../slave", {"alt": "de", "absolute": True}, "/de/projects/sklave/"),
        ("/projects/slave", {"alt": "de", "absolute": True}, "/de/projects/sklave/"),
        # Test external
        # Error if external set and no project URL
        (".", {"external": True}, NO_PROJECT_URL),
        (".", {"external": True, "base_url": "/content/"}, NO_PROJECT_URL),
        ("/projects/slave", {"alt": "de", "external": True}, NO_PROJECT_URL),
        # External ignored if absolute is set
        (".", {"absolute": True, "external": True}, "/projects/wolf/"),
        (
            ".",
            {"absolute": True, "external": True, "base_url": "/content/"},
            "/projects/wolf/",
        ),
        # Test base_url
        (".", {"base_url": "/"}, "projects/wolf/"),
        (".", {"alt": "de", "base_url": "/"}, "de/projects/wolf/"),
        (".", {"base_url": "/content/"}, "../projects/wolf/"),
        (".", {"absolute": True, "base_url": "/content/"}, "/projects/wolf/"),
        # No trailing slash
        ("/extra/file.ext", {}, "../../extra/file.ext"),
        ("/extra/file.ext", {"alt": "de", "absolute": True}, "/de/extra/file.ext"),
        # Custom slug
        ("/extra/slash-slug", {}, "../../extra/long/path/"),
        # Non-resolvable paths are silently treated as URL paths
        ("missing", {}, "missing"),
        ("./missing", {}, "missing"),
        ("../missing", {}, "../missing"),
        ("/missing", {}, "../../missing"),
        ("missing", {"alt": "de"}, "missing"),
        ("../missing", {"alt": "de"}, "../missing"),
        ("missing", {"absolute": True}, "/projects/wolf/missing"),
        ("missing", {"external": True}, NO_PROJECT_URL),
        ("missing", {"base_url": "/projects/"}, "wolf/missing"),
        # Disable resolution with '!' prefix
        ("!missing", {}, "missing"),
        ("!/extra/slash-slug", {}, "../../extra/slash-slug"),
        ("!", {}, "./"),
        ("!.", {}, "."),
        ("!../slave", {}, "../slave"),  # no trailing slash
        ("!../slave", {"alt": "de"}, "../slave"),  # alt ignored
        ("!/projects/slave", {}, "../slave"),
        ("!../slave", {"absolute": True}, "/projects/slave"),
        ("!../slave", {"external": True}, NO_PROJECT_URL),
        ("!../slave", {"absolute": True, "external": True}, "/projects/slave"),
        ("!../slave", {"base_url": "/"}, "projects/slave"),
        # Non-resolvable page falls back to URL-path resolution
        # With full outer product of resolve and strict_resolve options
        ("missing", {}, "missing"),
        ("missing", {"resolve": True}, "missing"),
        ("missing", {"resolve": False}, "missing"),
        ("missing", {"strict_resolve": False}, "missing"),
        ("missing", {"strict_resolve": True}, CAN_NOT_RESOLVE),
        ("missing", {"resolve": True, "strict_resolve": False}, "missing"),
        ("missing", {"resolve": True, "strict_resolve": True}, CAN_NOT_RESOLVE),
        ("missing", {"resolve": False, "strict_resolve": False}, "missing"),
        ("missing", {"resolve": False, "strict_resolve": True}, "missing"),
        # Resolve=True can not be used with '!' prefix
        ("!missing", {"resolve": True}, RESOLVE_INCOMPATIBLE_WITH_BANG),
        ("!missing", {"resolve": False}, "missing"),
        ("!missing", {"strict_resolve": True}, RESOLVE_INCOMPATIBLE_WITH_BANG),
        ("!missing", {"strict_resolve": False}, "missing"),
        ("!missing", {"resolve": False, "strict_resolve": True}, "missing"),
        # Resolvable page with full outer product of resolve and strict_resolve options
        ("/extra/slash-slug", {}, "../../extra/long/path/"),
        ("/extra/slash-slug", {"resolve": True}, "../../extra/long/path/"),
        ("/extra/slash-slug", {"resolve": False}, "../../extra/slash-slug"),
        ("/extra/slash-slug", {"strict_resolve": False}, "../../extra/long/path/"),
        ("/extra/slash-slug", {"strict_resolve": True}, "../../extra/long/path/"),
        (
            "/extra/slash-slug",
            {"resolve": True, "strict_resolve": False},
            "../../extra/long/path/",
        ),
        (
            "/extra/slash-slug",
            {"resolve": True, "strict_resolve": True},
            "../../extra/long/path/",
        ),
        (
            "/extra/slash-slug",
            {"resolve": False, "strict_resolve": False},
            "../../extra/slash-slug",
        ),
        (
            "/extra/slash-slug",
            {"resolve": False, "strict_resolve": True},
            "../../extra/slash-slug",
        ),
        # Alt ignored when resolve=False
        ("../slave", {"alt": "de", "resolve": False}, "../slave"),
        ("../slave", {"alt": "de", "resolve": True}, "../../de/projects/sklave/"),
        # Test anchor is preserved
        ("/projects/slave#anchor", {}, "../slave/#anchor"),
        ("missing#anchor", {}, "missing#anchor"),
        ("!missing#anchor", {}, "missing#anchor"),
        ("/extra#anchor", {"resolve": False}, "../../extra#anchor"),
        # Test anchor is preserved for non-resolved urls
        ("missing?q", {}, "missing?q"),
        ("!missing?q", {}, "missing?q"),
        ("/extra?q", {"resolve": False}, "../../extra?q"),
        # Test query and anchor are preserved for non-resolved urls
        ("missing?q#a", {}, "missing?q#a"),
        ("!missing?q#a", {}, "missing?q#a"),
        ("/extra?q#a", {"resolve": False}, "../../extra?q#a"),
        # Test explicit alt in query
        ("../slave?alt=de", {}, "../../de/projects/sklave/"),
        ("/extra?alt=de", {}, "../../de/extra/"),
        # Conflicting specifications for alt
        ("../slave?alt=de", {"alt": "en"}, CONFLICTING_ALT),
        # Unrecognized query params are discarded for resolved urls
        ("../slave?unrecognized=de", {}, "../slave/"),
        ("/extra?query#anchor", {}, "../../extra/#anchor"),
        # External urls
        ("http://example.org/slave", {}, "http://example.org/slave"),
        ("//example.com/?q#a", {}, "//example.com/?q#a"),
        ("some-scheme:foo", {}, "some-scheme:foo"),
    ],
    ids=_id_for_dict,
)
def test_url_to_all_params(pad, path, kwargs, expected):
    wolf_en = pad.get("/projects/wolf")
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as exc_info:
            wolf_en.url_to(path, **kwargs)
        assert re.search(str(expected), str(exc_info.value))
    else:
        assert wolf_en.url_to(path, **kwargs) == expected


def test_url_parse_virtual_path(pad):
    projects_de = pad.get("/projects@2", alt="de")
    paginated_de = pad.get("/extra/paginated@2", alt="de")
    paginated = pad.get("/extra/paginated@2")
    assert projects_de.url_path == "/de/projects/page/2/"
    assert paginated_de.url_path == "/de/extra/paginiert/page/2/"
    assert paginated.url_path == "/extra/paginated/page/2/"


def test_file_record_url(pad):
    record = pad.get("/extra/file.ext")
    assert record.url_path == "/extra/file.ext"


def test_url_with_slash_slug(pad):
    record = pad.get("/extra/slash-slug")
    assert record.url_path == "/extra/long/path/"


class DummyVirtualSource(VirtualSourceObject):
    url_path = None  # mask inherited property

    def __init__(self, record, url_path):
        super().__init__(record)
        self.url_path = url_path

    @property
    def url_content_path(self):
        # This is a url_content_path that is appropriate for a
        # source object that can contain other child source objects
        if self.url_path.endswith("/"):
            return self.url_path
        head, sep, tail = self.url_path.rpartition("/")
        return f"{head}{sep}_{tail}/"

    @property
    def path(self):
        return f"{self.record.path}@virtual"


@pytest.mark.parametrize(
    "path, url_path, expected",
    [
        ("/", "/path/virtual/", "../../"),
        ("/index.html", "/path/virtual/", "../../index.html"),
        ("rel", "/path/virtual/", "rel"),
        ("rel", "/path/virtual.html", "_virtual.html/rel"),
    ],
)
def test_url_from_virtual(pad, path, url_path, expected):
    virtual = DummyVirtualSource(pad.get("/extra"), url_path)
    assert virtual.url_to(path) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("a", "_file.ext/a"),
        ("/", "../"),
        ("/projects", "../projects/"),
    ],
)
def test_url_from_page_with_dotted_name(pad, path, expected):
    record = pad.get("/extra/file.ext")
    assert record.url_to(path) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("a", "a"),
        ("/", "../"),
        ("/projects", "../projects/"),
    ],
)
def test_url_from_attachment(pad, path, expected):
    record = pad.get("/extra/hello.txt")
    assert record.url_to(path) == expected


def test_url_from_attachment_issues_warning(pad):
    record = pad.get("/extra/hello.txt")
    with BufferReporter(pad.env) as reporter:
        assert record.url_to("a") == "a"
    message = next(filter(None, (extra.get("message") for _, extra in reporter.buffer)))
    assert re.match(r"(?i)Suspicious use of relative URL", message)



import inspect
from pathlib import Path

from lektor.project import Project


def test_Project_get_output_path(tmp_path: Path) -> None:
    project_file = tmp_path / "test.lektorproject"
    project_file.touch()
    project = Project.from_file(project_file)
    assert Path(project.get_output_path()).parts[-2:] == ("builds", project.id)


def test_Project_get_output_path_is_relative_to_project_file(tmp_path: Path) -> None:
    tree = tmp_path / "tree"
    tree.mkdir()
    project_file = tmp_path / "test.lektorproject"
    project_file.write_text(
        inspect.cleandoc(
            """[project]
            path = tree
            output_path = htdocs
            """
        )
    )

    project = Project.from_file(project_file)
    assert project.get_output_path() == str(tmp_path / "htdocs")



import os
import shutil

import pytest

from lektor.builder import Builder
from lektor.context import Context
from lektor.db import Database
from lektor.db import Siblings
from lektor.environment import Environment
from lektor.project import Project
from lektor.reporter import Reporter


@pytest.fixture(scope="function")
def pntest_project(tmp_path, data_path):
    src = data_path / "dependency-test-project"
    tmp_project = tmp_path / "project"
    shutil.copytree(src, tmp_project)
    return Project.from_path(tmp_project)


@pytest.fixture
def pntest_env(pntest_project, save_sys_path):
    return Environment(pntest_project)


@pytest.fixture
def pntest_pad(pntest_env):
    return Database(pntest_env).new_pad()


class DependencyReporter(Reporter):
    def __init__(self, *args, **kwargs):
        Reporter.__init__(self, *args, **kwargs)
        self.deps = {}

    def report_dependencies(self, dependencies):
        source_id = self.current_artifact.source_obj["_id"]
        row = self.deps.setdefault(source_id, set())
        for (
            _artifact_name,
            source_path,
            _mtime,
            _size,
            _checksum,
            _is_dir,
            _is_primary,
        ) in dependencies:
            row.add(source_path)

    @property
    def artifact_ids(self):
        return set(self.deps.keys())

    def clear(self):
        self.deps = {}


@pytest.fixture
def pntest_reporter(pntest_env):
    reporter = DependencyReporter(pntest_env)
    reporter.push()
    try:
        yield reporter
    finally:
        reporter.pop()


def test_prev_next_dependencies(tmp_path, pntest_env, pntest_reporter):
    env, reporter = pntest_env, pntest_reporter
    builder = Builder(env.new_pad(), tmp_path / "output")
    builder.build_all()

    # We start with posts 1, 2, and 4. Posts 1 and 2 depend on each other,
    # posts 2 and 3 depend on each other, but posts 1 and 3 are independent.
    assert "content/post2/contents.lr" in reporter.deps["post1"]
    assert "content/post3/contents.lr" not in reporter.deps["post1"]
    assert "/post1@siblings" in reporter.deps["post1"]

    assert "content/post1/contents.lr" in reporter.deps["post2"]
    assert "content/post4/contents.lr" in reporter.deps["post2"]
    assert "/post2@siblings" in reporter.deps["post2"]

    assert "content/post1/contents.lr" not in reporter.deps["post4"]
    assert "content/post2/contents.lr" in reporter.deps["post4"]
    assert "/post4@siblings" in reporter.deps["post4"]

    # Create post3, check that post2 and post4's dependencies are updated.
    post3_dir = os.path.join(env.project.tree, "content", "post3")

    os.makedirs(post3_dir)
    with open(os.path.join(post3_dir, "contents.lr"), "w+", encoding="utf-8"):
        pass

    reporter.clear()
    builder = Builder(env.new_pad(), builder.destination_path)
    builder.build_all()

    # post2, post3, and post4 had to be rebuilt, but not post1.
    assert {"post2", "post3", "post4"} == set(reporter.artifact_ids)

    # post2 depends on post3 now, not post4.
    assert "content/post3/contents.lr" in reporter.deps["post2"]
    assert "content/post4/contents.lr" not in reporter.deps["post2"]

    # post4 depends on post3 now, not post2.
    assert "content/post3/contents.lr" in reporter.deps["post4"]
    assert "content/post2/contents.lr" not in reporter.deps["post4"]


def test_prev_next_virtual_resolver(pntest_pad):
    with Context(pad=pntest_pad):
        siblings = pntest_pad.get("post2@siblings")
        assert isinstance(siblings, Siblings)
        assert siblings.path == "/post2@siblings"
        assert siblings.record["_id"] == "post2"
        assert siblings.prev_page["_id"] == "post1"
        assert siblings.next_page["_id"] == "post4"

        with pytest.raises(NotImplementedError):
            _ = siblings.url_path



import os
import textwrap

import pytest
from iniconfig import IniConfig
from inifile import IniFile

import lektor.quickstart
from lektor.cli import cli


@pytest.fixture(scope="session")
def can_symlink(tmp_path_factory):
    """Test whether current user has sufficient privileges to create symlinks."""
    if os.name == "nt":
        tmp_path = tmp_path_factory.mktemp("symlink-test")
        try:
            os.symlink("foo", tmp_path / "test")
        except OSError as exc:
            # Error Code 1314 - A required privilege is not held by the client
            # pylint: disable=no-member
            if exc.winerror != 1314:
                raise
            return False
    return True


@pytest.fixture
def default_author(mocker):
    author = "J. Random Hacker"
    mocker.patch.object(
        lektor.quickstart, "get_default_author", spec=True, return_value=author
    )
    return author


@pytest.fixture
def default_author_email(mocker):
    email = "jrh@example.org"
    mocker.patch.object(
        lektor.quickstart, "get_default_author_email", spec=True, return_value=email
    )
    return email


def _pack_lines(*lines: str) -> str:
    """Pack input lines into single string."""
    return "\n".join(lines) + "\n"


# new-plugin
def test_new_plugin(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin"],
        input=_pack_lines("Plugin Name", "", "Author Name", "author@email.com", "y"),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = os.path.join("packages", "plugin-name")
    assert set(os.listdir(path)) == {
        "lektor_plugin_name.py",
        "pyproject.toml",
        "setup.cfg",
        ".gitignore",
        "README.md",
    }

    # gitignore
    gitignore_expected = textwrap.dedent(
        """
        /dist
        /build
        __pycache__
        *.egg-info
    """
    ).strip()
    with open(os.path.join(path, ".gitignore"), encoding="utf-8") as f:
        gitignore_contents = f.read().strip()
    assert gitignore_contents == gitignore_expected

    # README.md
    readme_expected = textwrap.dedent(
        """
        # Plugin Name

        This is where a description of your plugin goes.
        Provide usage instructions here.
    """
    ).strip()
    with open(os.path.join(path, "README.md"), encoding="utf-8") as f:
        readme_contents = f.read().strip()
    assert readme_contents == readme_expected

    setup = IniConfig(os.path.join(path, "setup.cfg"))
    assert setup.get("metadata", "name") == "lektor-plugin-name"
    assert setup.get("metadata", "author") == "Author Name"
    assert setup.get("metadata", "author_email") == "author@email.com"
    assert setup.get("options", "py_modules") == "lektor_plugin_name"
    assert (
        setup.get("options.entry_points", "lektor.plugins")
        == "plugin-name = lektor_plugin_name:PluginNamePlugin"
    )

    # plugin.py
    plugin_expected = textwrap.dedent(
        """
        from lektor.pluginsystem import Plugin


        class PluginNamePlugin(Plugin):
            name = 'Plugin Name'
            description = "Add your description here."

            def on_process_template_context(self, context, **extra):
                def test_function():
                    return f"Value from plugin {self.name}"
                context["test_function"] = test_function
    """
    ).strip()
    with open(os.path.join(path, "lektor_plugin_name.py"), encoding="utf-8") as f:
        plugin_contents = f.read().strip()
    assert plugin_contents == plugin_expected


def test_new_plugin_abort_plugin_exists(project_cli_runner):
    path = "packages"
    os.mkdir(path)
    os.mkdir(os.path.join(path, "plugin-name"))
    input = "Plugin Name\n\nAuthor Name\nauthor@email.com\ny\n"
    result = project_cli_runner.invoke(cli, ["dev", "new-plugin"], input=input)
    assert "Aborted!" in result.output
    assert result.exit_code == 1


def test_new_plugin_abort_cancel(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin"],
        input=_pack_lines("Plugin Name", "", "Author Name", "author@email.com", "n"),
    )
    assert "Aborted!" in result.output
    assert result.exit_code == 1


def test_new_plugin_name_only(project_cli_runner, default_author, default_author_email):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin"],
        input=_pack_lines("Plugin Name", "", "", "", "y"),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = "packages"
    assert os.listdir(path) == ["plugin-name"]

    setup = IniConfig(os.path.join(path, "plugin-name", "setup.cfg"))
    assert setup.get("metadata", "author") == default_author
    assert setup.get("metadata", "author_email") == default_author_email


def test_new_plugin_name_param(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin", "plugin-name"],
        input=_pack_lines("", "Author Name", "author@email.com", "y"),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = "packages"
    assert os.listdir(path) == ["plugin-name"]


def test_new_plugin_path(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin"],
        input=_pack_lines(
            "Plugin Name", "path", "Author Name", "author@email.com", "y"
        ),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {
        "lektor_plugin_name.py",
        "pyproject.toml",
        "setup.cfg",
        ".gitignore",
        "README.md",
    }


def test_new_plugin_path_param(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin", "--path", "path"],
        input=_pack_lines("Plugin Name", "Author Name", "author@email.com", "y"),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {
        "lektor_plugin_name.py",
        "pyproject.toml",
        "setup.cfg",
        ".gitignore",
        "README.md",
    }


def test_new_plugin_path_and_name_params(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-plugin", "plugin-name", "--path", "path"],
        input=_pack_lines("Author Name", "author@email.com", "y"),
    )
    assert "Create Plugin?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {
        "lektor_plugin_name.py",
        "pyproject.toml",
        "setup.cfg",
        ".gitignore",
        "README.md",
    }


# new-theme
def test_new_theme(project_cli_runner, can_symlink):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme"],
        input=_pack_lines(
            "Lektor Theme Name", "", "Author Name", "author@email.com", "y"
        ),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = os.path.join("themes", "lektor-theme-name")
    assert set(os.listdir(path)) == {"example-site", "images", "README.md", "theme.ini"}
    assert set(os.listdir(os.path.join(path, "images"))) == {"homepage.png"}
    assert set(os.listdir(os.path.join(path, "example-site"))) == {
        "lektor-theme-name.lektorproject",
        "README.md",
        "themes",
    }
    if can_symlink:
        assert (
            os.readlink(os.path.join(path, "example-site/themes/lektor-theme-name"))
            == "../../../lektor-theme-name"
        )

    theme_inifile = IniFile(os.path.join(path, "theme.ini"))
    assert theme_inifile["theme.name"] == "Lektor Theme Name"
    assert theme_inifile["author.email"] == "author@email.com"
    assert theme_inifile["author.name"] == "Author Name"

    with open(os.path.join(path, "README.md"), encoding="utf-8") as f:
        readme_contents = f.read().strip()
    assert "Lektor Theme Name" in readme_contents


def test_new_theme_abort_theme_exists(project_cli_runner):
    path = "themes"
    os.mkdir(path)
    os.mkdir(os.path.join(path, "lektor-theme-name"))
    input = "Lektor Name\n\nAuthor Name\nauthor@email.com\ny\n"
    result = project_cli_runner.invoke(cli, ["dev", "new-theme"], input=input)
    assert "Aborted!" in result.output
    assert result.exit_code == 1


def test_new_theme_abort_cancel(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme"],
        input=_pack_lines("Theme Name", "", "Author Name", "author@email.com", "n"),
    )
    assert "Aborted!" in result.output
    assert result.exit_code == 1


def test_new_theme_name_only(
    project_cli_runner, can_symlink, default_author, default_author_email
):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme"],
        input=_pack_lines("Lektor Name Theme", "", "", "", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = os.path.join("themes", "lektor-theme-name")
    assert set(os.listdir(path)) == {"example-site", "images", "README.md", "theme.ini"}
    assert set(os.listdir(os.path.join(path, "images"))) == {"homepage.png"}
    assert set(os.listdir(os.path.join(path, "example-site"))) == {
        "lektor-theme-name.lektorproject",
        "README.md",
        "themes",
    }
    if can_symlink:
        assert (
            os.readlink(os.path.join(path, "example-site/themes/lektor-theme-name"))
            == "../../../lektor-theme-name"
        )

    theme_inifile = IniFile(os.path.join(path, "theme.ini"))
    assert theme_inifile["theme.name"] == "Lektor Name Theme"
    assert theme_inifile["author.email"] == default_author_email
    assert theme_inifile["author.name"] == default_author

    with open(os.path.join(path, "README.md"), encoding="utf-8") as f:
        readme_contents = f.read().strip()
    assert "Lektor Name Theme" in readme_contents


def test_new_theme_name_param(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme", "theme-name"],
        input=_pack_lines("", "Author Name", "author@email.com", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = "themes"
    assert os.listdir(path) == ["lektor-theme-name"]


def test_new_theme_path(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme"],
        input=_pack_lines("Theme Name", "path", "Author Name", "author@email.com", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {"example-site", "images", "theme.ini", "README.md"}


def test_new_theme_path_param(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme", "--path", "path"],
        input=_pack_lines("Theme Name", "Author Name", "author@email.com", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {"example-site", "images", "theme.ini", "README.md"}


def test_new_theme_path_and_name_params(project_cli_runner):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme", "theme-name", "--path", "path"],
        input=_pack_lines("Author Name", "author@email.com", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    path = "path"
    assert set(os.listdir(path)) == {"example-site", "images", "theme.ini", "README.md"}


@pytest.mark.parametrize(
    ("theme_name", "expected_id"),
    (
        ("Lektor New Theme", "lektor-theme-new"),
        ("Lektor Theme", "lektor-theme-theme"),
        ("New Theme", "lektor-theme-new"),
        ("New", "lektor-theme-new"),
        ("Theme", "lektor-theme-theme"),
        ("Lektor", "lektor-theme-lektor"),
        ("Lektor Theme New", "lektor-theme-new"),
    ),
)
@pytest.mark.usefixtures("default_author", "default_author_email")
def test_new_theme_varying_names(project_cli_runner, theme_name, expected_id):
    result = project_cli_runner.invoke(
        cli,
        ["dev", "new-theme"],
        input=_pack_lines(f"{theme_name}", "", "", "", "y"),
    )
    assert "Create Theme?" in result.output
    assert result.exit_code == 0
    assert expected_id in os.listdir("themes")
    path = os.path.join("themes", expected_id, "example-site")
    assert expected_id + ".lektorproject" in os.listdir(path)



"""MarkdownController implementation for mistune 0.x"""

import threading
from typing import ClassVar

import mistune  # type: ignore[import]

from lektor.markdown.controller import MarkdownController
from lektor.markdown.controller import Meta  # FIXME: move this?
from lektor.markdown.controller import RendererHelper
from lektor.sourceobj import SourceObject
from lektor.utils import deprecated


def escape(text: str) -> str:
    return mistune.escape(text, quote=True)


class ImprovedRenderer(
    # pylint: disable=no-member
    mistune.Renderer  # type: ignore[misc]
):
    lektor: ClassVar = RendererHelper()

    @property  # type: ignore[misc] # https://github.com/python/mypy/issues/1362
    @deprecated("Use ImprovedRenderer.lektor.record instead.", version="3.4.0")
    def record(self) -> SourceObject:
        return self.lektor.record

    @property  # type: ignore[misc]
    @deprecated("Use ImprovedRenderer.lektor.meta instead.", version="3.4.0")
    def meta(self) -> Meta:
        return self.lektor.meta

    def link(self, link: str, title: str | None, text: str) -> str:
        url = self.lektor.resolve_url(link)
        if not title:
            return f'<a href="{escape(url)}">{text}</a>'
        return f'<a href="{escape(url)}" title="{escape(title)}">{text}</a>'

    def image(self, src: str, title: str | None, text: str) -> str:
        url = escape(self.lektor.resolve_url(src))
        if not title:
            return f'<img src="{url}" alt="{escape(text)}">'
        return f'<img src="{url}" alt="{escape(text)}" title="{escape(title)}">'


class MarkdownConfig:
    def __init__(self) -> None:
        self.options = {
            "escape": False,
        }
        self.renderer_base = ImprovedRenderer
        self.renderer_mixins: list[type] = []

    def make_renderer(self) -> ImprovedRenderer:
        bases = tuple(self.renderer_mixins) + (self.renderer_base,)
        renderer_cls = type("renderer_cls", bases, {})
        return renderer_cls(**self.options)


class MarkdownController0(MarkdownController, threading.local):
    # NB: making this a threading.local means the results in the
    # cached_property MarkdownController.parser having a separate
    # value in each thread.
    #
    # We need that since the mistune 0.x parser is not thread-safe.

    def make_parser(self) -> mistune.Markdown:
        env = self.env
        cfg = MarkdownConfig()
        env.plugin_controller.emit("markdown-config", config=cfg)
        renderer = cfg.make_renderer()
        env.plugin_controller.emit(
            "markdown-lexer-config", config=cfg, renderer=renderer
        )
        # pylint: disable=unexpected-keyword-arg
        return mistune.Markdown(renderer, **cfg.options)



from __future__ import annotations

import os
import shutil
import sys
import threading
import warnings
from collections.abc import Callable
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path

import pytest
from watchfiles import Change

from lektor.environment import Environment
from lektor.project import Project
from lektor.watcher import watch_project
from lektor.watcher import WatchFilter


RunInThread = Callable[[Callable[[], None]], None]


@pytest.fixture
def run_in_thread() -> RunInThread:
    threads = []

    def run_thread(target: Callable[[], None]) -> None:
        t = threading.Thread(target=target)
        t.start()
        threads.append(t)

    try:
        yield run_thread
    finally:
        for t in threads:
            t.join(10.0)
        for t in threads:
            assert not t.is_alive()


@dataclass
class WatchResult:
    change_seen: bool = False

    def __bool__(self):
        return self.change_seen


@dataclass
class WatcherTest:
    env: Environment
    run_in_thread: RunInThread

    @contextmanager
    def __call__(
        self,
        timeout: float = 1.2,
    ) -> Generator[WatchResult]:
        """Run watch_project in a separate thread, wait for a file change event.

        This is a context manager that runs watch_project in a separate thread.  After
        the context exits, it will wait at most ``timeout`` seconds before returning.
        If a file system change is seen, it will return immediately.

        The context manager returns a WatchResult value.  After the context has been
        exited, the result will be True-ish if a file system change was noticed,
        False-ish otherwise.

        """
        if sys.platform == "darwin":
            self.macos_pause_for_calm()

        with self.watch(timeout) as change_seen:
            yield change_seen

    @contextmanager
    def watch(
        self,
        timeout: float,
    ) -> Generator[WatchResult]:
        """Run watch_project in a separate thread, wait for a file change event."""
        running = threading.Event()
        stop = threading.Event()
        changed = threading.Event()

        def run() -> None:
            watcher = watch_project(
                self.env, "non-existant-output-path", stop_event=stop
            )
            running.set()
            for _ in watcher:
                changed.set()
                return

        self.run_in_thread(run)
        result = WatchResult()
        running.wait()
        try:
            yield result
            result.change_seen = changed.wait(timeout)
        finally:
            stop.set()

    def macos_pause_for_calm(self) -> None:
        # Wait a bit for the dust to settle.
        # For whatever reason, on macOS, the watcher sometimes seems to return
        # filesystem events that happened shortly before it was started.
        for n in range(5):
            with self.watch(timeout=0.1) as change_seen:
                pass
            if not change_seen:
                break
            warnings.warn(f"macOS settle loop {n}: {change_seen}")  # noqa: B028


@pytest.fixture
def watcher_test(scratch_env: Environment, run_in_thread: RunInThread) -> WatcherTest:
    return WatcherTest(scratch_env, run_in_thread)


@pytest.fixture
def watched_path(scratch_env: Environment) -> Path:
    return Path(scratch_env.root_path)


def test_watcher_test(watcher_test: WatcherTest) -> None:
    with watcher_test(timeout=0.2) as change_seen:
        pass
    assert not change_seen


def test_sees_created_file(watcher_test: WatcherTest, watched_path: Path) -> None:
    with watcher_test() as change_seen:
        Path(watched_path, "created").touch()
    assert change_seen


def test_sees_deleted_file(watcher_test: WatcherTest, watched_path: Path) -> None:
    deleted_path = watched_path / "deleted"
    deleted_path.touch()

    with watcher_test() as change_seen:
        deleted_path.unlink()
    assert change_seen


def test_sees_modified_file(watcher_test: WatcherTest, watched_path: Path) -> None:
    modified_path = watched_path / "modified"
    modified_path.touch()

    with watcher_test() as change_seen:
        with modified_path.open("a") as fp:
            fp.write("addition")
    assert change_seen


def test_sees_file_moved_in(
    watcher_test: WatcherTest, watched_path: Path, tmp_path: Path
) -> None:
    orig_path = tmp_path / "orig_path"
    orig_path.touch()
    final_path = watched_path / "final_path"

    with watcher_test() as change_seen:
        orig_path.rename(final_path)
    assert change_seen


def test_sees_file_moved_out(
    watcher_test: WatcherTest, watched_path: Path, tmp_path: Path
) -> None:
    orig_path = watched_path / "orig_path"
    orig_path.touch()
    final_path = tmp_path / "final_path"

    with watcher_test() as change_seen:
        orig_path.rename(final_path)
    assert change_seen


def test_sees_deleted_directory(watcher_test: WatcherTest, watched_path: Path) -> None:
    # We only really care about deleted directories that contain at least a file.
    deleted_path = watched_path / "deleted"
    deleted_path.mkdir()
    watched_file = deleted_path / "file"
    watched_file.touch()

    with watcher_test() as change_seen:
        shutil.rmtree(deleted_path)
    assert change_seen


def test_sees_file_in_directory_moved_in(
    watcher_test: WatcherTest, watched_path: Path, tmp_path: Path
) -> None:
    # We only really care about directories that contain at least a file.
    orig_dir_path = tmp_path / "orig_dir_path"
    orig_dir_path.mkdir()
    Path(orig_dir_path, "file").touch()
    final_dir_path = watched_path / "final_dir_path"

    with watcher_test() as change_seen:
        orig_dir_path.rename(final_dir_path)
    assert change_seen


def test_sees_directory_moved_out(
    watcher_test: WatcherTest, watched_path: Path, tmp_path: Path
) -> None:
    # We only really care about directories that contain at least one file.
    orig_dir_path = watched_path / "orig_dir_path"
    orig_dir_path.mkdir()
    Path(orig_dir_path, "file").touch()
    final_dir_path = tmp_path / "final_dir_path"

    with watcher_test() as change_seen:
        orig_dir_path.rename(final_dir_path)
    assert change_seen


def test_ignores_opened_file(watcher_test: WatcherTest, watched_path: Path) -> None:
    file_path = watched_path / "file"
    file_path.touch()

    with watcher_test() as change_seen:
        with file_path.open() as fp:
            fp.read()
    assert not change_seen


@pytest.fixture(scope="session")
def watch_filter(project: Project) -> WatchFilter:
    env = Environment(project, load_plugins=False)
    return WatchFilter(env)


@pytest.mark.parametrize("path", [".dotfile", "webpack/node_modules"])
def test_WatchFilter_false(
    watch_filter: WatchFilter, path: str, project: Project
) -> None:
    abspath = os.path.abspath(os.path.join(project.tree, path))
    assert not watch_filter(Change.added, abspath)



import importlib
import tempfile
from urllib.parse import urlsplit

import pytest
from werkzeug import urls as werkzeug_urls

from lektor.compat import _CompatURL


@pytest.mark.parametrize(
    "name, expected_value, replacement",
    [
        (
            "TemporaryDirectory",
            tempfile.TemporaryDirectory,
            "tempfile.TemporaryDirectory",
        ),
        ("importlib_metadata", importlib.metadata, "importlib.metadata"),
    ],
)
def test_deprecated_attr(name, expected_value, replacement):
    lektor_compat = importlib.import_module("lektor.compat")
    with pytest.deprecated_call(match=f"use {replacement} instead") as warnings:
        value = getattr(lektor_compat, name)
    assert value is expected_value
    assert warnings[0].filename == __file__


def test_missing_attr():
    lektor_compat = importlib.import_module("lektor.compat")
    with pytest.raises(AttributeError):
        _ = lektor_compat.MISSING


def make_CompatURL(url: str) -> _CompatURL:
    return _CompatURL._make(urlsplit(url))


URL_MAKERS = [
    pytest.param(make_CompatURL, id="_CompatURL"),
    pytest.param(
        getattr(werkzeug_urls, "url_parse", None),
        id="werkzeug.urls.URL",
        marks=pytest.mark.skipif(
            not hasattr(werkzeug_urls, "url_parse"),
            reason="werkzeug.urls.url_parse is not available",
        ),
    ),
]


@pytest.fixture(params=URL_MAKERS)
def make_url(request):
    return request.param


pytestmark = pytest.mark.filterwarnings("ignore:'werkzeug:DeprecationWarning")


def test_compatURL_replace(make_url):
    url = make_url("http://example.org/foo")
    assert str(url.replace(scheme="https", path="bar")) == "https://example.org/bar"


def test_compatURL_host(make_url):
    assert make_url("http://bücher.example/foo").host == "bücher.example"


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http:/foo", [None]),
        ("http://bücher.example/foo", ["xn--bcher-kva.example"]),
        # werkzeug < 2.3 strips non-ascii
        ("http://bad\uffff.example.org", ["bad\uffff.example.org", "bad.example.org"]),
    ],
)
def test_compatURL_ascii_host(url, expected, make_url):
    assert make_url(url).ascii_host in expected


def test_compatURL_auth(make_url):
    assert make_url("http://u:p@example.org/").auth == "u:p"


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http://example.org/", None),
        ("http://a%20user@example.org/", "a user"),
        ("http://%C3%BCser:pw@example.org/", "üser"),
        ("http://%FCser:pw@example.org/", "üser"),  # latin1
    ],
)
def test_compatURL_username(url, expected, make_url):
    assert make_url(url).username == expected


def test_compatURL_raw_username(make_url):
    assert make_url("http://a%20user:p@example.org/").raw_username == "a%20user"


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http://user@example.org/", None),
        ("http://user:pass%20word@example.org/", "pass word"),
    ],
)
def test_compatURL_password(url, expected, make_url):
    assert make_url(url).password == expected


def test_compatURL_raw_password(make_url):
    assert make_url("http://u:pass%20word@example.org/").raw_password == "pass%20word"


def test_compatURL_decode_query(make_url):
    args = make_url("?a=b&c").decode_query()
    assert args["a"] == "b"
    assert args["c"] == ""


def test_compatURL_join(make_url):
    base = make_url("http://example.org/top/index.html")
    assert str(base.join("sibling.html")) == "http://example.org/top/sibling.html"


def test_compatURL_join_tuple(make_url):
    base = make_url("http://example.org/top/index.html")
    url = ("", "", "sibling.html", "q", "f")
    assert str(base.join(url)) == "http://example.org/top/sibling.html?q#f"


def test_compatURL_to_url(make_url):
    strval = "http://example.org/foo?q#a"
    url = make_url(strval)
    assert url.to_url() == strval
    assert str(url) == strval


@pytest.mark.filterwarnings("ignore:(?i).*werkzeug:DeprecationWarning")
def test_compatURL_to_uri_tuple(make_url):
    uri_tuple = make_url("http://example.org/fü/").to_uri_tuple()
    assert uri_tuple == ("http", "example.org", "/f%C3%BC/", "", "")


@pytest.mark.filterwarnings("ignore:(?i).*werkzeug:DeprecationWarning")
def test_compatURL_to_iri_tuple(make_url):
    iri_tuple = make_url("http://example.org/f%C3%BC/").to_iri_tuple()
    assert iri_tuple == ("http", "example.org", "/fü/", "", "")



import threading
from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from collections.abc import Hashable
from collections.abc import Mapping
from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Any
from typing import NamedTuple
from typing import Optional
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from werkzeug.utils import cached_property

from lektor.context import Context
from lektor.context import get_ctx
from lektor.sourceobj import SourceObject


if TYPE_CHECKING:  # pragma: no cover
    from lektor.environment import Environment


@dataclass
class _Threadlocal(threading.local):
    renderer_context: Optional["RendererContext"] = None


_threadlocal = _Threadlocal()

Meta = dict[str, Any]
FieldOptions = Mapping[str, str]


def require_ctx() -> Context:
    """Get Lektor build context, raising error if there is no current context."""
    ctx = get_ctx()
    if ctx is None:
        raise RuntimeError("Context is required for markdown rendering")
    return ctx


class RendererContext(NamedTuple):
    """Extra data used during Markdown rendering."""

    record: SourceObject | None
    meta: Meta
    field_options: FieldOptions

    def __enter__(self) -> "RendererContext":
        assert _threadlocal.renderer_context is None
        _threadlocal.renderer_context = self
        return self

    def __exit__(self, *__: Any) -> None:
        _threadlocal.renderer_context = None


def get_renderer_context() -> RendererContext:
    if _threadlocal.renderer_context is None:
        raise RuntimeError("RendererContext is required for markdown rendering")
    return _threadlocal.renderer_context


class RendererHelper:
    """Various helpers used by our markdown renderer subclasses."""

    @property
    def record(self) -> SourceObject | None:
        """The record that owns the markdown field being rendered.

        This is used as the base for resolving relative URLs in the Markdown text.
        """
        return get_renderer_context().record

    @property
    def meta(self) -> Meta:
        """The metadata for the current render.

        This is a dict and is used to return metadata from the
        rendering process.  Currently, Lektor itself never generates
        any metadata, but custom Lektor plugins can do so by updating
        this dict.

        Values inserted into this dict during the rendering process
        may be accessed in jinja templates via the ``.meta`` attribute
        of the _markdown_ field.
        """
        return get_renderer_context().meta

    @property
    def field_options(self) -> FieldOptions:
        """Field options.

        A mapping containing the options specified on the markdown field in
        the model.ini file.
        """
        return get_renderer_context().field_options

    @property
    def base_url(self) -> str:
        """Get current base_url from build context.

        The base URL of the artifact being built. This should start
        with a "/", however note that it is interpreted relative to
        any base_path configured for the project.

        """
        return require_ctx().base_url

    def resolve_url(self, url: str) -> str:
        """Resolve markdown link to a URL."""
        resolve_links = self.field_options.get("resolve_links")
        # Default is to resolve links to Lektor source objects when possible
        # This is a change from previous versions where we never resolved
        # links in Markdown.
        record = self.record
        if record is None:
            if resolve_links == "always":
                raise RuntimeError("A source object is required to resolve URLs")
            return url

        resolve = strict_resolve = None
        if resolve_links == "always":
            strict_resolve = True
        elif resolve_links == "never":
            # This is the old behavior, equivalent to '!' prefix
            resolve = False
        return self.record.url_to(
            url, base_url=self.base_url, resolve=resolve, strict_resolve=strict_resolve
        )


class UnknownPluginError(LookupError):
    """Exception raised when a mistune2 plugin name can not be resolved."""


class RenderResult(NamedTuple):
    html: str
    meta: Meta


class MarkdownController(ABC):
    def __init__(self, env: "Environment") -> None:
        self.env = env

    @abstractmethod
    def make_parser(self) -> Callable[[str], str]:  # () -> mistune.Mistune
        """Construct a mistune parser"""

    @cached_property
    def parser(self) -> Callable[[str], str]:  # () -> mistune.Mistune
        return self.make_parser()

    def get_cache_key(self) -> Hashable | None:
        """Get cache key.

        Identical keys guarantee that the rendered result for a given string,
        record, and set of field options will be identical.

        This method may return ``None`` to disable caching of results.
        """
        # pylint: disable=no-self-use
        ctx = get_ctx()
        if ctx is None:
            return None
        return ctx.base_url

    def render(
        self, source: str, record: SourceObject | None, field_options: FieldOptions
    ) -> RenderResult:
        """Render markdown string"""
        meta: Meta = {}
        self.env.plugin_controller.emit("markdown-meta-init", meta=meta, record=record)
        with RendererContext(record, meta, field_options):
            html = self.parser(source)
        self.env.plugin_controller.emit(
            "markdown-meta-postprocess", meta=meta, record=record
        )
        return RenderResult(html, meta)


class ControllerCache:
    """Helper for constructing MarkdownControllers thats ensures just one
    controller per Lektor Environment."""

    _cache: MutableMapping["Environment", MarkdownController]

    def __init__(self, factory: Callable[["Environment"], MarkdownController]):
        self.controller_class = factory
        self._cache = WeakKeyDictionary()

    def __call__(self, env: Optional["Environment"] = None) -> MarkdownController:
        """Get MarkdownController for environment.

        If no value is passed for env, the env for the current Lektor build context
        will be used.
        """
        if env is None:
            env = require_ctx().env
        try:
            return self._cache[env]
        except KeyError:
            return self._cache.setdefault(env, self.controller_class(env))



from os.path import dirname
from os.path import join

import pytest

from lektor.publisher import RsyncPublisher


def test_get_server(env):
    server = env.load_config().get_server("production")
    assert server.name == "Production"
    assert server.name_i18n["de"] == "Produktion"
    assert server.target == "rsync://myserver.com/path/to/website"
    assert server.extra == {"extra_field": "extra_value"}


def test_rsync_command_credentials(tmpdir, mocker, env):
    output_path = tmpdir.mkdir("output")
    publisher = RsyncPublisher(env, str(output_path))
    target_url = "http://example.com"
    credentials = {
        "username": "fakeuser",
        "password": "fakepass",
    }
    mock_popen = mocker.patch("lektor.publisher.portable_popen")
    with publisher.get_command(target_url, credentials):
        assert mock_popen.called
        assert mock_popen.call_args[0] == (
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                str(output_path) + "/",
                "fakeuser@example.com:/",
            ],
        )


output_path = join(dirname(__file__), "OUTPUT_PATH")


@pytest.mark.parametrize(
    "target_url,called_command",
    [
        (
            "http://example.com",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "http://fakeuser@example.com",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                str(output_path) + "/",
                "fakeuser@example.com:/",
            ],
        ),
        (
            "http://example.com?exclude=file",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--exclude",
                "file",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "http://example.com?exclude=file_one&exclude=file_two",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--exclude",
                "file_one",
                "--exclude",
                "file_two",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            """http://example.com?exclude='user's "special" file name'""",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--exclude",
                "'user's \"special\" file name'",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            'http://example.com?exclude="file name"',
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--exclude",
                '"file name"',
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "http://example.com?delete",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--delete-after",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "http://example.com?delete=yes",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                "--delete-after",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "http://example.com?delete=no",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                str(output_path) + "/",
                "example.com:/",
            ],
        ),
        (
            "file:///path/to/directory",
            [
                "rsync",
                "-rclzv",
                "--exclude=.lektor",
                str(output_path) + "/",
                "/path/to/directory/",
            ],
        ),
    ],
)
def test_rsync_publisher(target_url, called_command, tmpdir, mocker, env):
    publisher = RsyncPublisher(env, str(output_path))
    mock_popen = mocker.patch("lektor.publisher.portable_popen")
    with publisher.get_command(target_url, credentials=None):
        assert mock_popen.called
        assert mock_popen.call_args[0] == (called_command,)



from __future__ import annotations

import posixpath
from typing import TYPE_CHECKING
from urllib.parse import parse_qsl
from urllib.parse import urlsplit
from weakref import ref as weakref

from lektor.constants import PRIMARY_ALT
from lektor.context import ignore_url_unaffecting_dependencies
from lektor.reporter import reporter
from lektor.utils import is_path_child_of
from lektor.utils import join_path


if TYPE_CHECKING:
    from lektor.db import Pad


class SourceObject:
    source_classification = "generic"

    # We consider this class at least what public usage is to considered
    # to be from another place.
    __module__ = "db"

    def __init__(self, pad: Pad):
        self._pad = weakref(pad)

    @property
    def alt(self):
        """Returns the effective alt of this source object (unresolved)."""
        return PRIMARY_ALT

    @property
    def source_filename(self):
        """The primary source filename of this source object.

        In general, subclasses should implement/override ``iter_source_filenames``
        rather than this property.
        """
        source_filenames = self.iter_source_filenames()
        return next(iter(source_filenames), None)

    is_hidden = False
    is_discoverable = True

    @property
    def is_visible(self):
        """The negated version of :attr:`is_hidden`."""
        return not self.is_hidden

    @property
    def is_undiscoverable(self):
        """The negated version of :attr:`is_discoverable`."""
        return not self.is_discoverable

    def iter_source_filenames(self):
        """An iterable of the source filenames for this source object.

        The first returned filename should be the "primary" one.
        """
        # pylint: disable=no-self-use
        return ()

    @property
    def url_path(self):
        """The URL path of this source object if available."""
        raise NotImplementedError()

    @property
    def path(self):
        """Return the full path to the source object.  Not every source
        object actually has a path but source objects without paths need
        to subclass `VirtualSourceObject`.
        """
        return None

    @property
    def pad(self) -> Pad:
        """The associated pad of this source object."""
        rv = self._pad()
        if rv is not None:
            return rv
        raise AttributeError("The pad went away")

    def resolve_url_path(self, url_path):
        """Given a URL path as list this resolves the most appropriate
        direct child and returns the list of remaining items.  If no
        match can be found, the result is `None`.
        """
        if not url_path:
            return self
        return None

    def is_child_of(self, path, strict=False):
        """Checks if the current object is a child of the passed object
        or path.
        """
        if isinstance(path, SourceObject):
            path = path.path
        if self.path is None or path is None:
            return False
        return is_path_child_of(self.path, path, strict=strict)

    def url_to(
        self,
        path,  # : Union[str, "SourceObject", "SupportsUrlPath"]
        *,
        alt: str | None = None,
        absolute: bool | None = None,
        external: bool | None = None,
        base_url: str | None = None,
        resolve: bool | None = None,
        strict_resolve: bool | None = None,
    ) -> str:
        """Calculates the URL from the current source object to the given
        other source object.  Alternatively a path can also be provided
        instead of a source object.  If the path starts with a leading
        bang (``!``) then no resolving is performed.

        If a `base_url` is provided then it's used instead of the URL of
        the record itself.

        If path is a string and resolve=False is passed, then no attempt is
        made to resolve the path to a Lektor source object.

        If path is a string and strict_resolve=True is passed, then an exception
        is raised if the path can not be resolved to a Lektor source object.

        API CHANGE: It used to be (lektor <= 3.3.1) that if absolute was true-ish,
        then a url_path (URL path relative to the site's ``base_path`` was returned.
        This is changed so that now an absolute URL path is returned.
        """
        if base_url is None:
            base_url = self.url_path
        if absolute:
            # This sort of reproduces the old behaviour, where when
            # ``absolute`` was trueish, the "absolute" URL path
            # (relative to config.base_path) was returned, regardless
            # of the value of ``external``.
            external = False
        if resolve is None and strict_resolve:
            resolve = True

        if isinstance(path, SourceObject):
            # assert not isinstance(path, Asset)
            target = path
            if alt is not None and alt != target.alt:
                # NB: path.path includes page_num
                alt_target = self.pad.get(path.path, alt=alt, persist=False)
                if alt_target is not None:
                    target = alt_target
                # FIXME: issue warning or fail if cannot get correct alt?
            url_path = target.url_path
        elif hasattr(path, "url_path"):  # e.g. Thumbnail
            assert path.url_path.startswith("/")
            url_path = path.url_path
        elif path[:1] == "!":
            # XXX: error if used with explicit alt?
            if resolve:
                raise RuntimeError("Resolve=True is incompatible with '!' prefix.")
            url_path = _join_url_path(self, path[1:])
        elif resolve is not None and not resolve:
            # XXX: error if used with explicit alt?
            url_path = _join_url_path(self, path)
        else:
            with ignore_url_unaffecting_dependencies():
                return self._resolve_url(
                    path,
                    alt=alt,
                    absolute=absolute,
                    external=external,
                    base_url=base_url,
                    strict=strict_resolve,
                )

        return self.pad.make_url(url_path, base_url, absolute, external)

    def _resolve_url(
        self,
        _url: str,
        alt: str | None,
        absolute: bool | None,
        external: bool | None,
        base_url: str | None,
        strict: bool | None,
    ) -> str:
        """Resolve (possibly relative) URL or db path to URL."""
        url = urlsplit(_url)
        if url.scheme or url.netloc:
            resolved = url
        else:
            # Interpret path as (possibly relative) db-path
            dbpath = join_path(self.path, url.path)
            params = dict(parse_qsl(url.query, keep_blank_values=False))
            query_alt = params.get("alt")
            # XXX: support page_num in query, too?
            if not alt:
                alt = query_alt or self.alt
            elif query_alt and query_alt != alt:
                raise RuntimeError("Conflicting values for alt.")
            target = self.pad.get(dbpath, alt=alt)
            if target is not None:
                url_path = target.url_path
                query = ""
            elif strict:
                raise RuntimeError(f"Can not resolve link {_url!r}")
            else:
                # Fall back to interpreting path as (possibly relative) URL path
                url_path = _join_url_path(self, url.path)
                query = url.query

            result = self.pad.make_url(
                url_path, absolute=absolute, external=external, base_url=base_url
            )
            resolved = urlsplit(result)._replace(query=query, fragment=url.fragment)
        return resolved.geturl()

    @property
    def url_content_path(self):
        """URL path to the directory that contains children of this source object.

        For container types, the record's ``url_content_path`` is often
        the same as its ``url_path``. The exception to this is when
        the page's slug contains a dot (".").
        See https://www.getlektor.com/docs/content/urls/#content-below-dotted-slugs

        The ``url_content_path`` should be ``None`` for attachments and other
        SourceObject types that can not contain child source objects.
        """
        return None


def _join_url_path(source, path):
    """Join possibly relative url path relative to source.url_content_path."""
    if posixpath.isabs(path):
        return path
    content_path = source.url_content_path
    if content_path is None:
        # Source is not a container type (e.g. it is an attachment).
        # Punt and treat path as relative to the source's containing directory.
        content_path = posixpath.dirname(source.url_path) or "/"
        reporter.report_generic(
            f"Suspicious use of relative URL {path!r} "
            f"from non-container source {source!r}"
        )
    return posixpath.join(content_path, path)


class DBSourceObject(SourceObject):
    """This is the base class for objects that live in the lektor db.

    I.e. this is the type of object returned by pad.get().

    """

    @property
    def path(self):
        """Return the full database path to the source object.

        All DBSourceObjects must have paths.
        """
        raise NotImplementedError()

    # XXX: move SourceObject.url_to here?

    def __eq__(self, other):
        if other is self:
            return True  # optimization
        if other.__class__ is not self.__class__:
            return False  # optimization
        return (
            other.alt == self.alt and other.path == self.path and other.pad == self.pad
        )

    def __hash__(self):
        return hash((self.path, self.alt))


class VirtualSourceObject(DBSourceObject):
    """Virtual source objects live below a parent record but do not
    originate from the source tree with a separate file.
    """

    def __init__(self, record):
        super().__init__(record.pad)
        self.record = record

    @property
    def path(self):
        raise NotImplementedError()

    def get_mtime(self, path_cache):
        # pylint: disable=no-self-use
        return None

    def get_checksum(self, path_cache):
        # pylint: disable=no-self-use
        return None

    @property
    def parent(self):
        return self.record

    @property
    def alt(self):
        return self.record.alt

    def iter_source_filenames(self):
        # This is a default.  However, if artifacts produced from a
        # particular virtual source type do not explicitly vary with
        # the parent record, it may make sense to override this to
        # return an empty (or some other) list of file names.
        return self.record.iter_source_filenames()



"""Thumbnail generation."""

from __future__ import annotations

import dataclasses
import io
import math
import posixpath
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Any
from typing import ClassVar
from typing import Final
from typing import NamedTuple
from typing import TYPE_CHECKING

import PIL.Image
import PIL.ImageCms

from ..utils import get_dependent_url
from ._compat import PILLOW_VERSION_INFO
from ._compat import Transpose  # PIL.Image.Transpose
from .image_info import get_image_info
from .image_info import get_image_orientation
from .image_info import SvgImageInfo
from .image_info import TiffOrientation
from .image_info import UnknownImageInfo


if TYPE_CHECKING:
    from _typeshed import SupportsRead

    from lektor.builder import Artifact
    from lektor.context import Context


class ThumbnailMode(Enum):
    FIT = "fit"
    CROP = "crop"
    STRETCH = "stretch"

    DEFAULT = "fit"


class _FormatInfo:
    format: ClassVar[str]
    default_save_params: ClassVar[dict[str, Any]] = {}
    extensions: ClassVar[Sequence[str]]

    @classmethod
    def get_save_params(cls, thumbnail_params: ThumbnailParams) -> dict[str, Any]:
        """Compute kwargs to be passed to Image.save() when writing the thumbnail."""
        params = dict(cls.default_save_params)
        params.update(cls._extra_save_params(thumbnail_params))
        params["format"] = cls.format
        return params

    @classmethod
    def get_thumbnail_tag(cls, thumbnail_params: ThumbnailParams) -> str:
        """Get a string which serializes the thumbnail_params.

        This is value is used as a suffix when generating the file name for the
        thumbnail.
        """
        width, height = thumbnail_params.size
        bits = [f"{width}x{height}"]
        if thumbnail_params.crop:
            bits.append("crop")
        bits.extend(cls._extra_tag_bits(thumbnail_params))
        return "_".join(bits)

    @classmethod
    def get_ext(cls, proposed_ext: str | None = None) -> str:
        """Get file extension suitable for image format.

        If proposed_ext is an acceptable extension for the format, return that.
        Otherwise return the default extension for the format.
        """
        if proposed_ext is not None and proposed_ext.lower() in cls.extensions:
            return proposed_ext
        return cls.extensions[0]

    @staticmethod
    def _extra_save_params(
        thumbnail_params: ThumbnailParams,
    ) -> Mapping[str, Any] | Iterable[tuple[str, Any]]:
        return {}

    @staticmethod
    def _extra_tag_bits(thumbnail_params: ThumbnailParams) -> Iterable[str]:
        return ()


class _GifFormatInfo(_FormatInfo):
    format = "GIF"
    extensions = (".gif",)


class _PngFormatInfo(_FormatInfo):
    format = "PNG"
    default_save_params = {"compress_level": 7}
    extensions = (".png",)

    @staticmethod
    def _extra_save_params(
        thumbnail_params: ThumbnailParams,
    ) -> Iterator[tuple[str, Any]]:
        quality = thumbnail_params.quality
        if quality is not None:
            yield "compress_level", min(9, max(0, quality // 10))

    @classmethod
    def _extra_tag_bits(cls, thumbnail_params: ThumbnailParams) -> Iterable[str]:
        for key, value in cls._extra_save_params(thumbnail_params):
            assert key == "compress_level"
            yield f"q{value}"


class _JpegFormatInfo(_FormatInfo):
    format = "JPEG"
    default_save_params = {"quality": 85}
    extensions = (".jpeg", ".jpg")

    @staticmethod
    def _extra_save_params(
        thumbnail_params: ThumbnailParams,
    ) -> Iterator[tuple[str, Any]]:
        quality = thumbnail_params.quality
        if quality is not None:
            yield "quality", quality

    @classmethod
    def _extra_tag_bits(cls, thumbnail_params: ThumbnailParams) -> Iterable[str]:
        for key, value in cls._extra_save_params(thumbnail_params):
            assert key == "quality"
            yield f"q{value}"


class ImageSize(NamedTuple):
    width: int
    height: int


@dataclasses.dataclass
class ThumbnailParams:
    """Encapsulates the parameters necessary to generate a thumbnail."""

    size: ImageSize
    format: str
    quality: int | None = None
    crop: bool = False

    def __post_init__(self) -> None:
        format = self.format.upper()
        for format_info_cls in _FormatInfo.__subclasses__():
            if format_info_cls.format == format:
                break
        else:
            raise ValueError(f"unrecognized format ({self.format!r})")
        self.format_info = format_info_cls

    def get_save_params(self) -> Mapping[str, Any]:
        """Get kwargs to pass to Image.save() when writing the thumbnail."""
        return self.format_info.get_save_params(self)

    def get_ext(self, proposed_ext: str | None = None) -> str:
        """Get file extension for thumbnail.

        If proposed_ext is an acceptable extension for the thumbnail, return that.
        Otherwise return the default extension for the thumbnail format.
        """
        return self.format_info.get_ext(proposed_ext)

    def get_tag(self) -> str:
        """Get a string which serializes the thumbnail_params.

        This is value is used as a suffix when generating the file name for the
        thumbnail.
        """
        return self.format_info.get_thumbnail_tag(self)


def _scale(x: int, num: float, denom: float) -> int:
    """Compute x * num / denom, rounded to integer.

    ``x``, ``num``, and ``denom`` should all be positive.

    Rounds 0.5 up to be consistent with imagemagick.

    """
    if isinstance(num, int) and isinstance(denom, int):
        # If all arguments are integers, carry out the computation using integer math to
        # ensure that 0.5 rounds up.
        return (x * num + denom // 2) // denom
    # If floats are involved, we do our best to round 0.5 up, but loss of precision
    # involved in floating point math makes the idea of "exactly" 0.5 a little fuzzy.
    return math.trunc((x * num + denom / 2) // denom)


def compute_dimensions(
    width: int | None, height: int | None, source_width: float, source_height: float
) -> ImageSize:
    """Compute "fit"-mode dimensions of thumbnail.

    Returns the maximum size of a thumbnail with that has (nearly) the same aspect ratio
    as the source and whose maximum size is set by ``width`` and ``height``.

    One, but not both, of ``width`` or ``height`` can be ``None``.
    """
    if width is None and height is None:
        raise ValueError("width and height may not both be None")
    if width is not None:
        size = ImageSize(width, _scale(width, source_height, source_width))
    if height is not None and (width is None or height < size.height):
        size = ImageSize(_scale(height, source_width, source_height), height)
    return size


class CropBox(NamedTuple):
    # field names taken from
    # https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.crop
    left: int
    upper: int
    right: int
    lower: int


def _compute_cropbox(size: ImageSize, source_width: int, source_height: int) -> CropBox:
    """Compute "crop"-mode crop-box to be applied to the source image before
    it is scaled to the final thumbnail dimensions.

    """
    use_width = min(source_width, _scale(source_height, size.width, size.height))
    use_height = min(source_height, _scale(source_width, size.height, size.width))
    crop_l = (source_width - use_width) // 2
    crop_t = (source_height - use_height) // 2
    return CropBox(crop_l, crop_t, crop_l + use_width, crop_t + use_height)


SRGB_PROFILE: Final = PIL.ImageCms.createProfile("sRGB")
SRGB_PROFILE_BYTES: Final = PIL.ImageCms.ImageCmsProfile(SRGB_PROFILE).tobytes()


def _get_icc_transform(
    icc_profile: bytes, inMode: str, outMode: str
) -> PIL.ImageCms.ImageCmsTransform:
    """Construct ICC transform mapping icc_profile to sRGB."""
    profile = PIL.ImageCms.getOpenProfile(io.BytesIO(icc_profile))
    transform = PIL.ImageCms.buildTransform(profile, SRGB_PROFILE, inMode, outMode)
    assert isinstance(transform, PIL.ImageCms.ImageCmsTransform)
    return transform


def _convert_to_rgb(image: PIL.Image.Image) -> PIL.Image.Image:
    # Ensure image is RGB before scaling
    targetMode = "RGBA" if image.mode.upper().endswith("A") else "RGB"
    if image.mode != targetMode:
        icc_profile = image.info.get("icc_profile")
        if icc_profile is not None:
            icc_transform = _get_icc_transform(icc_profile, image.mode, targetMode)
            image = PIL.ImageCms.applyTransform(image, icc_transform)
            del image.info["icc_profile"]
        else:
            image = image.convert(targetMode)
    return image


def _convert_icc_profile_to_srgb(image: PIL.Image.Image) -> None:
    """Convert image from embedded ICC profile to sRGB.

    The image is modified **in place**.

    After conversion, any embedded color profile is removed. (The default color
    space for the web is "sRGB", so we don't need to embed it.)
    """
    # XXX: The old imagemagick code (which ran `convert` with `-strip -colorspace sRGB`)
    # did not attempt any colorspace conversion.  It simply stripped and ignored any
    # color profile in the input image (causing the resulting thumbnail to be
    # interpreted as if it were in sRGB even though its not.)
    #
    # Here we attempt to convert from any embedded colorspace in the source image
    # to sRGB.
    #
    # Note: _convert_to_rgb may have already done this if the original image was not in
    # RGB(A) mode.  In that case it will have removed the icc_profile from the images
    # .info dict.
    #
    # XXX: There seems to be no real way to compare to color profiles to see whether
    # they are the same.  It's not even clear that all "sRGB" profiles are really the
    # same.  (See
    # https://ninedegreesbelow.com/photography/srgb-profile-comparison.html.) So we
    # always convert if the image has a color profile.
    #
    icc_profile = image.info.get("icc_profile")
    if icc_profile is not None and icc_profile != SRGB_PROFILE_BYTES:
        icc_transform = _get_icc_transform(icc_profile, image.mode, image.mode)
        PIL.ImageCms.applyTransform(image, icc_transform, inPlace=True)
        del image.info["icc_profile"]


_TRANSPOSE_FOR_ORIENTATION: Final[Mapping[TiffOrientation, int]] = {
    TiffOrientation.TOPRIGHT: Transpose.FLIP_LEFT_RIGHT,
    TiffOrientation.BOTRIGHT: Transpose.ROTATE_180,
    TiffOrientation.BOTLEFT: Transpose.FLIP_TOP_BOTTOM,
    TiffOrientation.LEFTTOP: Transpose.TRANSPOSE,
    TiffOrientation.RIGHTTOP: Transpose.ROTATE_270,
    TiffOrientation.RIGHTBOT: Transpose.TRANSVERSE,
    TiffOrientation.LEFTBOT: Transpose.ROTATE_90,
}


def _auto_orient_image(image: PIL.Image.Image) -> PIL.Image.Image:
    """Transpose image as indicated by the Exif Orientation tag.

    We only do this for JPEG images.  See _get_image_orientation for notes on why.

    """
    orientation = get_image_orientation(image)
    if orientation in _TRANSPOSE_FOR_ORIENTATION:
        image = image.transpose(
            _TRANSPOSE_FOR_ORIENTATION[orientation]  # type: ignore[arg-type]
        )
    return image


def _create_thumbnail(
    image: PIL.Image.Image, params: ThumbnailParams
) -> PIL.Image.Image:
    # XXX: There is an Image.thumbnail() method that can be significantly faster at
    # down-scaling than Image.resize() in some particular cases. Perhaps we want to use
    # that. (Image.thumbnail() never upscales, so we can only use it when down-scaling.)
    #
    # Image.thumbnail *only* has a possible advantage when down-scaling JPEG images,
    # where it configures the image loader to help with the down-scaling.  (With other
    # image types, .thumbnail just loads the image normally then uses .resize.)
    #
    # Some tests, downscaling a 5MB 4032x3024 JPEG, using:
    #
    #     python -m timeit -s "from PIL import Image" \
    #       "im = Image.open('in.jpg'); im.thumbnail((W,H)); im.save('out.jpg')"
    # or
    #     python -m timeit -s "from PIL import Image" \
    #       "Image.open('in.jpg').resize((W,H)[,reducing_gap=3]).save('out.jpg')"
    #
    #         WxH    |  .resize()  |  .thumbnail()  |  .resize(reducing_gap=3)
    #     ===========|=============|================|===========================
    #      1024x768  |  117 msec   |   115 msec     |  130 msec
    #       512x384  |  105 msec   |    51 msec     |   82 msec
    #       256x192  |  103 msec   |    33 msec     |   63 msec
    #       120x90   |  100 msec   |    22 msec     |   60 msec
    #         4x3    |   88 msec   |    22 msec     |   58 msec
    #
    # Thumbnail() by default uses reducing_gap=2.
    #
    # The big wins for .thumbnail() appear to come when downscaling by a factor of ~8 or
    # more.

    # Ensure image is in RGB (or RGBA) mode before scaling
    image = _convert_to_rgb(image)

    # transpose according to EXIF Orientation
    image = _auto_orient_image(image)

    # resize
    resize_params: dict[str, Any] = {"reducing_gap": 3.0}
    if params.crop:
        resize_params["box"] = _compute_cropbox(params.size, image.width, image.height)
    if PILLOW_VERSION_INFO < (7, 0):
        del resize_params["reducing_gap"]  # not supported in older Pillow
    thumbnail = image.resize(params.size, **resize_params)

    # Convert from any embedded ICC color profile to sRGB.
    _convert_icc_profile_to_srgb(thumbnail)

    # Do not propagate comment tag, or XMP data to thumbnail
    thumbnail.info.pop("comment", None)
    thumbnail.info.pop("xmp", None)

    return thumbnail


def _create_artifact(
    source_image: str | Path | SupportsRead[bytes],
    thumbnail_params: ThumbnailParams,
    artifact: Artifact,
) -> None:
    """Create artifact by computing thumbnail for source image."""
    with PIL.Image.open(source_image) as image:
        thumbnail = _create_thumbnail(image, thumbnail_params)
        save_params = thumbnail_params.get_save_params()
    with artifact.open("wb") as fp:
        thumbnail.save(fp, **save_params)


def _get_thumbnail_url_path(
    source_url_path: str, thumbnail_params: ThumbnailParams
) -> str:
    source_ext = posixpath.splitext(source_url_path)[1]
    # leave ext unchanged from source if valid for the thumbnail format
    ext = thumbnail_params.get_ext(source_ext)
    suffix = thumbnail_params.get_tag()
    return get_dependent_url(  # type: ignore[no-any-return]
        source_url_path, suffix, ext=ext
    )


def make_image_thumbnail(
    ctx: Context,
    source_image: str | Path,
    source_url_path: str,
    *,
    width: int | None = None,
    height: int | None = None,
    mode: ThumbnailMode = ThumbnailMode.DEFAULT,
    upscale: bool | None = None,
    quality: int | None = None,
) -> Thumbnail:
    """Helper method that can create thumbnails from within the build process
    of an artifact.
    """
    image_info = get_image_info(source_image)
    if isinstance(image_info, UnknownImageInfo):
        raise RuntimeError("Cannot process unknown images")

    if mode == ThumbnailMode.FIT:
        if width is None and height is None:
            raise ValueError("Must specify at least one of width or height.")
        if image_info.width is None or image_info.height is None:
            assert isinstance(image_info, SvgImageInfo)
            raise ValueError("Cannot determine aspect ratio of SVG image.")
        if upscale is None:
            upscale = False
        size = compute_dimensions(width, height, image_info.width, image_info.height)
    else:
        if width is None or height is None:
            raise ValueError(
                f'"{mode.value}" mode requires both `width` and `height` '
                "to be specified."
            )
        if upscale is None:
            upscale = True
        size = ImageSize(width, height)

    # If we are dealing with an actual svg image, we do not actually
    # resize anything, we just return it. This is not ideal but it's
    # better than outright failing.
    if isinstance(image_info, SvgImageInfo):
        # XXX: Since we don't always know the original dimensions,
        # we currently omit the upscaling check for SVG images.
        return Thumbnail(source_url_path, size.width, size.height)

    would_upscale = size.width > image_info.width or size.height > image_info.height
    if would_upscale and not upscale:
        return Thumbnail(source_url_path, image_info.width, image_info.height)

    thumbnail_params = ThumbnailParams(
        size=size,
        format=image_info.format.upper(),
        quality=quality,
        crop=mode == ThumbnailMode.CROP,
    )
    dst_url_path = _get_thumbnail_url_path(source_url_path, thumbnail_params)

    ctx.add_sub_artifact(
        artifact_name=dst_url_path,
        sources=[source_image],
        build_func=partial(_create_artifact, source_image, thumbnail_params),
    )

    return Thumbnail(dst_url_path, size.width, size.height)


@dataclasses.dataclass(frozen=True)
class Thumbnail:
    """Holds information about a thumbnail."""

    url_path: str
    width: int
    height: int

    def __str__(self) -> str:
        return posixpath.basename(self.url_path)



"""A hatch build-hook that builds the front-end js/css at PEP 517 build time.

By default, when building the dist, if the built frontend code is not
present in the source directory, it will be built.  (Node must be installed,
or this will fail.)

Environment Variables
---------------------

There are a couple of environment variables that can be used to provide
additional control over the building of the frontend code.

- ``HATCH_BUILD_NO_HOOKS=true``: skip building the frontend code
- ``HATCH_BUILD_CLEAN=true``: force (re)building of the frontend code

"""

import shutil
import subprocess
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


# path to frontend source
FRONTEND = "frontend"

# path to compiled JS entry point
APP_JS = "lektor/admin/static/app.js"


class FrontendBuildHook(BuildHookInterface):
    """Hatching build hook to compile our frontend JS."""

    def clean(self, versions):
        """Called at the beginning of each PEP 517 build, if HATCH_BUILD_CLEAN is set.

        This implementation deletes any compiled frontend code that may be present
        in the source tree.
        """
        app = self.app
        root = self.root

        if not Path(root, FRONTEND).is_dir():
            app.display_info(
                "frontend source missing, skipping cleaning compiled output"
            )
            return

        output_path = Path(root, APP_JS).parent
        app.display_info(f"cleaning {output_path.relative_to(root)}")
        if output_path.is_dir():
            shutil.rmtree(output_path)

    def initialize(self, version, build_data):
        """Hook called before each package build.

        This implementation builds the compiled frontend source, but only
        if it is not already present in the source tree.

        Node (and npm) must be installed or this step will fail.
        """
        app = self.app
        root = self.root

        if Path(root, APP_JS).is_file():
            app.display_info(f"{APP_JS} exists, skipping frontend build")
            return

        try:
            proc = subprocess.run(
                "npm -v", capture_output=True, text=True, shell=True, check=True
            )
        except subprocess.CalledProcessError as exc:
            app.abort(f"{exc.cmd!r} failed (is node/npm installed?)")
        app.display_info(f"found npm version {proc.stdout.strip()}")

        frontend = Path(root, FRONTEND)
        if not frontend.is_dir():
            app.abort("frontend source is missing. can not build frontend")

        app.display_info("npm install")
        subprocess.run("npm install", cwd=frontend, shell=True, check=True)
        app.display_info("npm run build")
        subprocess.run("npm run build", cwd=frontend, shell=True, check=True)
        app.display_success("built frontend static files")



import importlib

import lektor.i18n


EXPECTED_LANGS = {
    "ca",
    "en",
    "fr",
    "ja",
    "nl",
    "pt",
    "tr",
    "de",
    "es",
    "it",
    "ko",
    "pl",
    "ru",
    "zh",
}


def test_translations_loaded():
    known_langs = set(lektor.i18n.KNOWN_LANGUAGES)
    assert known_langs == EXPECTED_LANGS


def test_loading_i18n_triggers_no_warnings(recwarn):
    importlib.reload(lektor.i18n)
    for warning in recwarn.list:
        print(warning)  # debugging: display warnings on stdout
    assert len(recwarn) == 0
